import sys

from .space_delimited import SpaceDelimited

# No stemmer

# Copied from https://meta.wikimedia.org/wiki/?oldid=13987775
stopwords = [
    "aasta", "aastal", "aastani", "aastast", "aastat", "aastatel", "aeg",
    "aga", "ainult", "ajal", "ajalugu", "alates", "all", "alla", "allikad",
    "ameerika", "and", "aprill", "artikkel", "artiklit", "asub", "august",
    "autor", "category", "class", "com", "commons", "coordinate",
    "detsember", "di", "eest", "eesti", "eestis", "ega", "ehk", "ei", "elu",
    "enam", "enne", "esimene", "esimese", "est", "euroopa", "ew", "file",
    "hiljem", "htm", "html", "http", "ide", "iga", "ii", "il", "ile", "image",
    "in", "index", "inglise", "inimesed", "ist", "jaanuar", "jpg", "juba",
    "juuli", "juuni", "juures", "jäi", "järgi", "järjesta", "kaks", "kas",
    "kategooria", "keel", "keeles", "keeletoimeta", "kes", "kirjandus",
    "koduleht", "kogu", "kohta", "kokku", "kolm", "koos", "korda", "kui",
    "kuid", "kuna", "kuni", "kus", "kuu", "kõige", "kõik", "le", "left",
    "liit", "link", "linna", "lisa", "lisaks", "lk", "läbi", "lõuna",
    "maailma", "maakonna", "mida", "mille", "mis", "mitte", "märts", "nad",
    "nagu", "name", "need", "neid", "neist", "nende", "news", "ng", "nii",
    "nime", "nimi", "ning", "november", "nr", "ns", "näiteks", "of",
    "oktoober", "old", "ole", "oli", "olid", "olla", "olnud", "oma", "on",
    "org", "osa", "palju", "pdf", "peale", "php", "pildi", "pilt", "pisi",
    "png", "pole", "poolt", "px", "pärast", "põhja", "ref", "region", "right",
    "riigi", "riik", "rohkem", "rootsi", "räägib", "saab", "saanud", "sai",
    "saksa", "saksamaa", "sama", "samal", "samuti", "seal", "seda", "see",
    "selle", "sellest", "seotud", "september", "sest", "siis", "siiski",
    "small", "sse", "surnud", "suur", "svg", "sündinud", "tagasi", "tallinn",
    "tallinna", "tartu", "teda", "teine", "teise", "teiste", "tema", "the",
    "thumb", "toimeta", "toimetaaeg", "tuntud", "type", "tõttu", "umbes",
    "usa", "uus", "vaata", "vahel", "vaid", "vald", "vana", "varem", "vastu",
    "veebruar", "veel", "vene", "venemaa", "viide", "viited", "von", "väga",
    "välislingid", "välja", "või", "võib", "www", "ära", "ühe", "üks", "üle",
    "ülikool",
]

try:
    import enchant
    dictionary = enchant.Dict("et")
except enchant.errors.DictNotFoundError:
    raise ImportError("No enchant-compatible dictionary found for 'et'.  " +
                      "Consider installing 'myspell-et'.")

badwords = [
    r"pede",
    r"loll(akas|id)?",
    r"t(s|š)mir",
    r"pers+e(s)?",
    r"munn(i|id)?",
    r"kaka(junn)?",
    r"noku",
    r"pede(d|kas)?",
    r"vitt(u)?",
    r"pask",
    r"homo(d)?",
    r"puts(i)?",
    r"pederast(id)?",
    r"sitt(a)?",
    r"lits(id)?",
    r"pihku",
    r"türa",
    r"jobu",
    r"sita(ne|junn)?",
    r"vitupea",
    r"neeger",
    r"nigga",
    r"junn(i)?",
    r"kepp(is|i|ida)?",
    r"tuss"
]
informals = [
    r"lol(l)?"
    r"yolo",
    r"jou",
    r"raisk",
    r"tere",
    r"suht",
    r"haha",
    r"tsau",
    r"tegelt",
    r"sakib",
    r"noob",
    r"hahaha",
    r"räme",
    r"ilge",
    r"lahe",
    r"hmm"
]


sys.modules[__name__] = SpaceDelimited(
    __name__,
    doc="""
estonian
=========

revision
--------
.. autoattribute:: revision.words
.. autoattribute:: revision.content_words
.. autoattribute:: revision.badwords
.. autoattribute:: revision.misspellings
.. autoattribute:: revision.informals
.. autoattribute:: revision.infonoise

parent_revision
---------------
.. autoattribute:: parent_revision.words
.. autoattribute:: parent_revision.content_words
.. autoattribute:: parent_revision.badwords
.. autoattribute:: parent_revision.misspellings
.. autoattribute:: parent_revision.informals

diff
----
.. autoattribute:: diff.words_added
.. autoattribute:: diff.words_removed
.. autoattribute:: diff.badwords_added
.. autoattribute:: diff.badwords_removed
.. autoattribute:: diff.misspellings_added
.. autoattribute:: diff.misspellings_removed
.. autoattribute:: diff.informals_added
.. autoattribute:: diff.informals_removed
    """,
    badwords=badwords,
    dictionary=dictionary,
    informals=informals,
    stopwords=stopwords
)



from ...datasources import diff
from ...features import Feature
from ..meta.regex_extractors import SegmentRegexExtractor
from ..meta.token_filter import TokenFilter
from .util import token_is_word


class Diff:

    def __init__(self, language):
        self.language = language
        self.prefix = language.__name__ + ".diff."

        self.added_words_list = TokenFilter(
            self.prefix + "words",
            diff.added_tokens,
            token_is_word
        )
        """
        Returns a list of all word tokens added in this revision.
        """

        self.words_added = Feature(
            self.prefix + "words_added", len,
            returns=int,
            depends_on=[self.added_words_list]
        )
        """
        Returns a count of words added.
        """

        self.removed_words_list = TokenFilter(
            self.prefix + "words",
            diff.removed_tokens,
            token_is_word
        )
        """
        Returns a list of all word tokens removed in this revision.
        """

        self.words_removed = Feature(
            self.prefix + "removed_words", len,
            returns=int,
            depends_on=[self.removed_words_list]
        )
        """
        Returns a count of words removed.
        """

        if language.resources.badwords:
            self.added_badwords_list = SegmentRegexExtractor(
                self.prefix + "added_badwords",
                diff.added_segments,
                language.resources.badwords
            )
            """
            Returns a list of badwords added
            """

            self.removed_badwords_list = SegmentRegexExtractor(
                self.prefix + "removed_badwords",
                diff.removed_segments,
                language.resources.badwords
            )
            """
            Returns a list of badwords removed
            """

            self.badwords_added = Feature(
                self.prefix + "badwords_added", len,
                returns=int,
                depends_on=[self.added_badwords_list]
            )
            """
            Returns a count of badwords added
            """

            self.badwords_removed = Feature(
                self.prefix + "badwords_removed", len,
                returns=int,
                depends_on=[self.removed_badwords_list]
            )
            """
            Returns a count of badwords removed
            """

        if language.resources.informals:
            self.added_informals_list = SegmentRegexExtractor(
                self.prefix + "added_informals",
                diff.added_segments,
                language.resources.informals
            )
            """
            Returns a list of informal words added
            """

            self.removed_informals_list = SegmentRegexExtractor(
                self.prefix + "removed_informals",
                diff.removed_segments,
                language.resources.informals
            )
            """
            Retuns a list of informal words removed
            """

            self.informals_added = Feature(
                self.prefix + "informals_added", len,
                returns=int,
                depends_on=[self.added_informals_list]
            )
            """
            Returns a count of informal words added
            """

            self.informals_removed = Feature(
                self.prefix + "informals_removed", len,
                returns=int,
                depends_on=[self.removed_informals_list]
            )
            """
            Returns a count of informal words removed
            """

        if language.resources.dictionary:

            self.added_misspellings_list = TokenFilter(
                self.prefix + "added_misspellings",
                self.added_words_list,
                self.not_in_dictionary
            )
            """
            Retuns a list of misspellings added
            """

            self.removed_misspellings_list = TokenFilter(
                self.prefix + "removed_misspellings",
                self.removed_words_list,
                self.not_in_dictionary
            )
            """
            Returns a list of misspellings removed
            """

            self.misspellings_added = Feature(
                self.prefix + "misspellings_added", len,
                returns=int,
                depends_on=[self.added_misspellings_list]
            )
            """
            Returns a count of the number of misspellings added
            """

            self.misspellings_removed = Feature(
                self.prefix + "misspellings_removed", len,
                returns=int,
                depends_on=[self.removed_misspellings_list]
            )
            """
            Returns a list of the number of misspellings removed
            """

    def in_dictionary(self, word):
        return self.language.resources.dictionary.check(str(word))

    def not_in_dictionary(self, word):
        return not self.language.resources.dictionary.check(str(word))



"""
This module implements a set of :class:`revscoring.Language`
-- collections of features that are language specific.

languages
+++++++++
.. automodule:: revscoring.languages.dutch

.. automodule:: revscoring.languages.english

.. automodule:: revscoring.languages.estonian

.. automodule:: revscoring.languages.french

.. automodule:: revscoring.languages.german

.. automodule:: revscoring.languages.hebrew

.. automodule:: revscoring.languages.indonesian

.. automodule:: revscoring.languages.italian

.. automodule:: revscoring.languages.persian

.. automodule:: revscoring.languages.portuguese

.. automodule:: revscoring.languages.spanish

.. automodule:: revscoring.languages.turkish

.. automodule:: revscoring.languages.ukrainian

.. automodule:: revscoring.languages.vietnamese

Base classes
++++++++++++
.. automodule:: revscoring.languages.language

.. automodule:: revscoring.languages.space_delimited

"""
from .language import Language

__all__ = [Language]



from ...datasources import Datasource


class Constant(Datasource):

    def __init__(self, name, value):
        self.value = value
        super().__init__(name, self.process)

    def process(self):
        return self.value



import sys

from .space_delimited import SpaceDelimited

try:
    from nltk.stem.snowball import SnowballStemmer
    stemmer = SnowballStemmer("italian")
except ValueError:
    raise ImportError("Could not load stemmer for {0}. ".format(__name__))

try:
    from nltk.corpus import stopwords as nltk_stopwords
    stopwords = set(nltk_stopwords.words('italian'))
except LookupError:
    raise ImportError("Could not load stopwords for {0}. ".format(__name__) +
                      "You may need to install the nltk 'stopwords' " +
                      "corpora.  See http://www.nltk.org/data.html")

try:
    import enchant
    dictionary = enchant.Dict("it")
except enchant.errors.DictNotFoundError:
    raise ImportError("No enchant-compatible dictionary found for 'it'.  " +
                      "Consider installing 'myspell-it'.")

badwords = [
    r"anale",
    r"ano",
    r"bastard[io]",
    r"buttana",
    r"cac(a|are|ata)",
    r"cacc(a|he|ol[ae])",
    r"caga(re|t[aeo])?",
    r"cavolat[ae]",
    r"cazz(at[ae]|[io]|on[ei])?",
    r"cesso",
    r"ciccione",
    r"ciuccia",
    r"coglion[ei]",
    r"cojone",
    r"cretin[io]",
    r"culo",
    r"deficent[ei]",
    r"fancazzista",
    r"fanculo",
    r"fi[cg](a|he|o)",
    r"fott(e|iti|uto)",
    r"fregna",
    r"froci(o|one)?",
    r"gnocca",
    r"idiota",
    r"incul(are|ato|o)",
    r"kazzo",
    r"maiala",
    r"merd(a|accia|e|os[oa])?",
    r"mignotta",
    r"min(ch|k)ia",
    r"pen[ei]",
    r"pip+[iae]",
    r"pirla",
    r"piscio",
    r"pisell(ino|o|one)",
    r"pompin[io]",
    r"porc[ao]",
    r"porno",
    r"pupu",
    r"putta(n[ea]|nella|niere)?",
    r"puzza(no|te|va)?", "puzz(i|olente|one)",
    r"ricchione",
    r"sb[ou]rr[ao]",
    r"scem[aio]",
    r"schifo(s[ao])?",
    r"scopa(re|t[aeo]|va)",
    r"scor+egg(ia|e)",
    r"seghe",
    r"sfigat[io]",
    r"siffredi",
    r"skifo",
    r"sperma",
    r"stocazzo",
    r"stronza(t[ae])?",
    r"stronz[io]",
    r"stupido",
    r"suca",  r"succhia",
    r"suka",
    r"tette",
    r"troi[ae]",
    r"trombare",
    r"vaf+anculo",
    r"znccn[kl]",
    r"zoccola"
]

informals = [
    "(ah)+a?",
    "amo",
    "avete",
    "banana",
    "bla",
    "brutt[io]",
    "cavolo",
    "chiamo",
    "ciao+",
    "cmq",
    "consigliamo",
    "corsivo",
    "dovete",
    "fermatevi",
    "frega",
    "grassetto",
    "(ha)+",
    "inserisci",
    "intestazione",
    "leggete",
    "lol",
    "mamma",
    "nascondi",
    "perche",
    "potete",
    "raga",
    "sapete",
    "scrivete",
    "scusate",
    "siete",
    "sto",
    "tua",
    "volete"
]


sys.modules[__name__] = SpaceDelimited(
    __name__,
    doc="""
italian
=======

revision
--------
.. autoattribute:: revision.words
.. autoattribute:: revision.content_words
.. autoattribute:: revision.badwords
.. autoattribute:: revision.misspellings
.. autoattribute:: revision.informals
.. autoattribute:: revision.infonoise

parent_revision
---------------
.. autoattribute:: parent_revision.words
.. autoattribute:: parent_revision.content_words
.. autoattribute:: parent_revision.badwords
.. autoattribute:: parent_revision.misspellings
.. autoattribute:: parent_revision.informals
.. autoattribute:: parent_revision.infonoise

diff
----
.. autoattribute:: diff.words_added
.. autoattribute:: diff.words_removed
.. autoattribute:: diff.badwords_added
.. autoattribute:: diff.badwords_removed
.. autoattribute:: diff.misspellings_added
.. autoattribute:: diff.misspellings_removed
.. autoattribute:: diff.informals_added
.. autoattribute:: diff.informals_removed
    """,
    badwords=badwords,
    dictionary=dictionary,
    informals=informals,
    stemmer=stemmer,
    stopwords=stopwords
)



from collections import namedtuple

from ..language import Language
from .diff import Diff
from .parent_revision import ParentRevision
from .revision import Revision

Resources = namedtuple(
    "Resources",
    ['badwords', 'dictionary', 'informals', 'stemmer', 'stopwords']
)


class SpaceDelimited(Language):
    """
    Implements a set of useful features for a space-delimited class of
    language.

    Features are made available depending on the arguments provided to the
    constructor.  Reference the language implementation for available features.

    :Parameters:
        name : str
            The name of the language.  Not that this name will be used when
            checking for equality between languages.
        badwords : `list` of `str`
            A list of regexes to be used to match badwords.  Note that word
            boundary characters will be placed around your regex.
            E.g. `r"\\\\bbadword\\\\b"`
        dictionary : :class:`enchant.Dict`
            An enchant dictionary for use in looking up words
        informals : `list` of `str`
            A list of regexes to be used to match informal words.  Like
            badwords, these regexes will be wrapped in word boundary
            characters.
        stemmer : :class:`nltk.stem.api.StemmerI`
            An nltk-based stemmer to be used to remove morphological affixes
            from words.
        stopwords : `set` of `str`
            A set of common words to use when processing the informational
            content of text.  See `stopwords` in `nltk.corpus`.
    """
    def __init__(self, name, doc=None, badwords=None, dictionary=None,
                 informals=None, stemmer=None, stopwords=None):
        super().__init__(name, doc=doc)
        self.prefix = name
        stopwords = set(stopwords) if stopwords is not None else None
        self.resources = Resources(badwords, dictionary, informals, stemmer,
                                   stopwords)
        """
        The set of resources that were made available to the constructor
        """

        self.diff = Diff(self)
        """
        A module containing language-specific, diff-based features
        """

        self.parent_revision = ParentRevision(self)
        """
        A module containing language-specific, parent revision-based features
        """

        self.revision = Revision(self)
        """
        A module containing language-specific, revision-based features
        """



import logging
import time

from deltas import segment_matcher

from . import parent_revision, revision
from .datasource import Datasource

logger = logging.getLogger(__name__)


def process_operations(a, b):
    start = time.time()
    operations = [op for op in segment_matcher.diff(a, b)]
    logger.debug("diff() of {0} and {1} tokens took {2} seconds."
                 .format(len(a), len(b), time.time() - start))

    return operations, a, b

operations = Datasource("diff.operations", process_operations,
                        depends_on=[parent_revision.tokens, revision.tokens])
"""
Returns a tuple that describes the difference between the parent revision text
and the current revision's text.

The tuple contains three fields:

* operations: `list` of :class:`deltas.Operation`
* A tokens: `list` of `str`
* B tokens: `list` of `str`
"""


def process_added_tokens(diff_operations):

    operations, a, b = diff_operations

    return [t for op in operations
            if op.name == "insert"
            for t in b[op.b1:op.b2]]

added_tokens = Datasource("diff.added_tokens", process_added_tokens,
                          depends_on=[operations])
"""
Returns a list of all tokens added in this revision.
"""


def process_removed_tokens(diff_operations):

    operations, a, b = diff_operations

    return [t for op in operations
            if op.name == "delete"
            for t in a[op.a1:op.a2]]

removed_tokens = Datasource("removed_tokens", process_removed_tokens,
                            depends_on=[operations])
"""
Returns a list of all tokens removed in this revision.
"""


def process_added_segments(diff_operations):
    operations, a, b = diff_operations

    return ["".join(b[op.b1:op.b2])
            for op in operations
            if op.name == "insert"]

added_segments = Datasource("diff.added_segments", process_added_segments,
                            depends_on=[operations])
"""
Returns a list of all contiguous segments of tokens added in this revision.
"""


def process_removed_segments(revision_diff):
    operations, a, b = revision_diff

    return ["".join(a[op.a1:op.a2])
            for op in operations
            if op.name == "delete"]

removed_segments = Datasource("diff.removed_segments",
                              process_removed_segments,
                              depends_on=[operations])
"""
Returns a list of all contiguous segments of tokens removed in this revision.
"""



import pickle

from nose.tools import eq_

from .. import french
from ...datasources import revision
from ...dependencies import solve
from .util import compare_extraction

BAD = [
    "con",
    "fesse",
    "foutre",
    "merde", "merdee",
    "merdique",
    "prostituee", "prostitue",
    "putain", "putes",
    "salop",
    "stupide"
]

OTHER = [
    "connection", "fitness", "le"
]


def test_badwords():
    compare_extraction(french.revision.badwords_list, BAD, OTHER)


def test_revision():
    # Words
    cache = {revision.text: "Wikipédia est un projet d’encyclopédie."}
    eq_(solve(french.revision.words_list, cache=cache),
        ["Wikipédia", "est", "un", "projet", "d’encyclopédie"])

    # Misspellings
    cache = {revision.text: 'Est un projet principe du worngly. <td>'}
    eq_(solve(french.revision.misspellings_list, cache=cache), ["worngly"])

    # Infonoise
    cache = {revision.text: "Est un projet principe."}
    eq_(solve(french.revision.infonoise, cache=cache), 13/19)


def test_pickling():

    eq_(french, pickle.loads(pickle.dumps(french)))



import sys

from .space_delimited import SpaceDelimited

try:
    import enchant
    dictionary = enchant.Dict("pt")
except enchant.errors.DictNotFoundError:
    raise ImportError("No enchant-compatible dictionary found for 'pt'.  " +
                      "Consider installing 'myspell-pt'.")

try:
    from nltk.stem.snowball import SnowballStemmer
    stemmer = SnowballStemmer("portuguese")
except ValueError:
    raise ImportError("Could not load stemmer for {0}. ".format(__name__))

try:
    from nltk.corpus import stopwords as nltk_stopwords
    stopwords = set(nltk_stopwords.words('portuguese'))
except LookupError:
    raise ImportError("Could not load stopwords for {0}. ".format(__name__) +
                      "You may need to install the nltk 'stopwords' " +
                      "corpora.  See http://www.nltk.org/data.html")


badwords = [
    r"babaca",  # douchebag
    r"bixa",  # ???
    r"boiola", r"boiolas",  # gay man
    r"boquete",  # blowjob
    r"bosta",  # shit
    r"bucet(a|inha)s?",  # pussy (vagina)
    r"bund(a|inha)s?",  # ass
    r"burr[ao]s?",  # donkey/jackass
    r"cacete",  # bludgeon
    r"cag([ao]|ad[ao]|and[ao]|aneir[ao]|ar|ou)s?",  # shit
    r"cara(i|io|lho)s?",  # fuck
    r"chat[ao]",  # boring
    r"(ch|x)up[ao](r|va|u)?s?",  # blow me
    r"cocô",  # poo
    r"comi",  # eat ???
    r"cona(ssa)?s?",  # cunt
    r"cuz([aã]o|inho)",  # asshole
    r"doido",  # crazy
    r"fed(e|ido)",  # stinks/stinky
    r"feia",  # ugly
    r"fendi",  # ???
    r"foda[sr]?", r"fude[sr]?",  # fuck
    r"gostos[aã][os]?", r"gostoso",  # yummy ???
    r"idiotas?",  # idiot
    r"loka", r"loko",  # crazy
    r"maconheiro",  # bothead
    r"mafia",  # mafia
    r"maldizentes",  # slanderers
    r"mecos",  # cum ???
    r"mentir(a|os[oa])s?",  # lie/liar
    r"merdas?",  # shit
    r"noob",  # noob
    r"ot[áa]rios?",  # sucker
    r"pariu",  # to give birth ???
    r"pategos",  # hick / yokel
    r"peid(a|ar|o|ei)s?",  # fart
    r"pênis",  # penis
    r"pilas?",  # dick
    r"piroca",  # dick
    r"poha",  # ???
    r"porcaria", r"porno",  # filth/porn
    r"porra",  # cum
    r"pum",  # fart
    r"punhet(a|eiro)",  # jack off / masturbate
    r"put(o|a|aria|eiro|inha)s?",  # bitch/hooker
    r"safado",  # shameless
    r"tesão",  # turn-on / horny
    r"tran[sz]ar",  # sex
    r"tr(eta|oxa)",  # bullshit
    r"vadia",  # bitch
    r"viad(agem|ão|inho|o)s?",  # gay person ("fucker")
    r"xixi"  # pee
]

informals = [
    r"adoro",  # love
    r"aki",  # ???
    r"amo",  # master
    r"(b+l+a+h*)+",  # bla, blah, bbblllaaaahhhhhblah
    r"carambas?",  # OMG
    r"coco",  # coconut
    r"copie[im]",  # I copied
    r"delicia",  # delicious
    r"editei",  # edited
    r"enfiar?",  # to stick (up one's ass)
    r"entao",  # then
    r"estrag(ar|uem)",  # spoiled / ruined
    r"fixe",  # cool
    r"gajo",  # dude
    r"h[aiou](h[aeiou])*", r"h[e](h[aeiou])+",  # hi, ha, hehe, hohoho
    r"k+",  # k, kkkkkkkkkkkkkkk
    r"lindo",  # pretty
    r"l+([uo]+l+)+",  # lol, LOLOLOL, LLLLoOOoLLL
    r"mae",  # mom
    r"mto",  # very
    r"naum",  # no (slang)
    r"n[óo]is",  # it's us (slang)
    r"odeio",  # hate
    r"oi+",  # hi
    r"ol[aá]",  # hello
    r"ratas?",  # "rat" -- a snitch
    r"(rs)+",  # lol
    r"tava",  # was / were (slang)
    r"tbm",  # also (slang)
    r"vao",  # vain
    r"vcs", r"voce", r"voces",  # you
    r"xau"  # bye
]


sys.modules[__name__] = SpaceDelimited(
    __name__,
    doc="""
portuguese
==========

revision
--------
.. autoattribute:: revision.words
.. autoattribute:: revision.content_words
.. autoattribute:: revision.badwords
.. autoattribute:: revision.misspellings
.. autoattribute:: revision.informals
.. autoattribute:: revision.infonoise

parent_revision
---------------
.. autoattribute:: parent_revision.words
.. autoattribute:: parent_revision.content_words
.. autoattribute:: parent_revision.badwords
.. autoattribute:: parent_revision.misspellings
.. autoattribute:: parent_revision.informals
.. autoattribute:: parent_revision.infonoise

diff
----
.. autoattribute:: diff.words_added
.. autoattribute:: diff.words_removed
.. autoattribute:: diff.badwords_added
.. autoattribute:: diff.badwords_removed
.. autoattribute:: diff.misspellings_added
.. autoattribute:: diff.misspellings_removed
.. autoattribute:: diff.informals_added
.. autoattribute:: diff.informals_removed
    """,
    badwords=badwords,
    informals=informals,
    dictionary=dictionary,
    stemmer=stemmer,
    stopwords=stopwords
)
"""
portuguese
"""



import pickle

from nose.tools import eq_

from .. import portuguese
from ...datasources import revision
from ...dependencies import solve
from .util import compare_extraction

BAD = [
    "babaca",
    "bixa",
    "boiola", "boiolas",
    "boquete",
    "bosta",
    "buceta", "bucetas", "bucetinha",
    "bunda", "bundinha",
    "burra", "burro", "burros",
    "cacete",
    "caga", "cagada", "cagado", "cagando", "caganeira", "cagar", "cagou",
    "carai", "caraio", "caralho",
    "chata", "chato",
    "chupa", "chupar", "chupava", "chupo", "chupou", "xupa",
    "cocô",
    "comi",
    "cona",
    "cuzao", "cuzão", "cuzinho",
    "doido",
    "fede", "fedido",
    "feia",
    "fendi",
    "foda", "fodas", "fude", "fuder",
    "gostosa", "gostosão", "gostosas", "gostoso",
    "idiota", "idiotas",
    "loka", "loko",
    "maconheiro",
    "mafia",
    "maldizentes",
    "mecos",
    "mentira", "mentiroso", "mentirosa", "mentirosos", "mentirosas",
    "merda",   "merdas",
    "noob",
    "otario", "otário", "otarios",
    "pariu",
    "pategos",
    "peida", "peidar", "peidei",  "peido", "peidos",
    "pênis",
    "pila", "pilas",
    "piroca",
    "poha",
    "porcaria", "porno",
    "porra",
    "pum",
    "punheta", "punheteiro",
    "puta", "putaria", "putas", "puteiro", "putinha",
    "puto", "putos",
    "safado",
    "tesão",
    "transar", "tranzar",
    "treta", "troxa",
    "vadia",
    "viadagem", "viadão",  "viadinho", "viadinhos",
    "viado", "viados",
    "xixi",
]

INFORMAL = [
    "adoro",
    "aki",
    "amo",
    "bla", "blablabla", "bbblllaaaahhhhhblah",
    "coco",
    "copiei", "copiem",
    "delicia",
    "editei",
    "enfia", "enfiar",
    "entao",
    "estraguem",
    "fixe",
    "gajo",
    "haha", "hahaha", "hehe", "hehehe",
    "kkk", "kkkk", "kkkkk", "kkkkkk", "kkkkkkk",
    "lindo",
    "lol",
    "mae",
    "mto",
    "naum",
    "nois",
    "odeio",
    "oi", "oiiiiiiiiii",
    "ola", "olá",
    "rata", "ratas",
    "rs", "rsrsrs",
    "tava",
    "tbm",
    "vao",
    "vcs", "voce", "voces",
    "xau"
]

OTHER = [
    """
    A batalha de Hastings foi travada em 14 de outubro de 1066 entre o exército
    franco-normando do duque Guilherme II da Normandia (r. 1035–1087) e um
    exército inglês sob o rei anglo-saxão Haroldo II (r. 1066), durante a
    conquista normanda da Inglaterra. Ocorreu cerca de 11 quilômetros a
    noroeste de Hastings, perto da atual cidade de Battle, em Sussex Oriental,
    e teve como resultado uma decisiva vitória normanda.
    """,
    "arvere"
]


def test_badwords():
    compare_extraction(portuguese.revision.badwords_list, BAD, OTHER)


def test_informals():
    compare_extraction(portuguese.revision.informals_list, INFORMAL, OTHER)


def test_revision():
    # Words
    cache = {revision.text: "A haver, rebeliões: e m80 da Normandia."}
    eq_(solve(portuguese.revision.words_list, cache=cache),
        ["A", "haver", "rebeliões", "e", "m80", "da", "Normandia"])

    # Misspellings
    cache = {revision.text: 'O número de vítimas é difícil worngly. <td>'}
    eq_(solve(portuguese.revision.misspellings_list, cache=cache), ["worngly"])

    # Infonoise
    cache = {revision.text: "Esta a o corrida!"}
    eq_(solve(portuguese.revision.infonoise, cache=cache), 4/13)


def test_pickling():

    eq_(portuguese, pickle.loads(pickle.dumps(portuguese)))



"""
.. autoclass:: RevisionMetadata

.. autoclass:: UserInfo
"""
from mwtypes import Timestamp


class RevisionMetadata:
    """
    Represents a revision's metadata.
    """
    __slots__ = ('rev_id', 'parent_id', 'user_text', 'user_id', 'timestamp',
                 'comment', 'page_id', 'page_namespace', 'page_title', 'bytes',
                 'minor')

    def __init__(self, rev_id=None, parent_id=None, user_text=None,
                 user_id=None, timestamp=None, comment=None, page_id=None,
                 page_namespace=None, page_title=None, bytes=None, minor=None):
        self.rev_id = int(rev_id) if rev_id is not None else None
        self.parent_id = int(parent_id) if parent_id is not None else None
        self.user_text = str(user_text) if user_text is not None else None
        self.user_id = int(user_id) if user_id is not None else None
        self.timestamp = Timestamp(timestamp) \
            if timestamp is not None else None
        self.comment = str(comment) if comment is not None else None
        self.page_id = int(page_id) if page_id is not None else None
        self.page_namespace = int(page_namespace) \
            if page_namespace is not None else None
        self.page_title = str(page_title) if page_title is not None else None
        self.bytes = int(bytes) if bytes is not None else None
        self.minor = bool(minor)


class UserInfo:
    """
    Represents information about a user.
    """
    __slots__ = ('id', 'name', 'editcount', 'registration',
                 'groups', 'implicitgroups', 'emailable',
                 'gender', 'block_id', 'blocked_by',
                 'blocked_by_id', 'blocked_timestamp', 'block_reason',
                 'block_expiry')

    def __init__(self, id=None, name=None, editcount=None, registration=None,
                 groups=None, implicitgroups=None, emailable=None,
                 gender=None, block_id=None, blocked_by=None,
                 blocked_by_id=None, blocked_timestamp=None, block_reason=None,
                 block_expiry=None):
        self.id = int(id) if id is not None else None
        self.name = str(name) if name is not None else None
        self.editcount = int(editcount) if editcount is not None else None
        self.registration = Timestamp(registration) \
            if registration is not None else None
        self.groups = groups or []
        self.implicitgroups = implicitgroups or []
        self.emailable = bool(emailable)
        self.gender = str(gender) if gender is not None else None
        self.block_id = int(block_id) if block_id is not None else None
        self.blocked_by = str(blocked_by) if blocked_by is not None else None
        self.blocked_by_id = int(blocked_by_id) \
            if blocked_by_id is not None else None
        self.blocked_timestamp = Timestamp(blocked_timestamp) \
            if blocked_timestamp is not None else None
        self.block_reason = str(block_reason) \
            if block_reason is not None else None
        self.block_expiry = str(block_expiry) \
            if block_expiry is not None else None



"""
``revscoring extract_features -h``
::

    Adds features to a set of labeled revisions.

    Reads a TSV file of <rev_id>\t<label> pairs and replaces the
    <rev_id> field with the extracted feature values.

    Input: <rev_id>[TAB]<label>

    Output: <feature_1>[TAB]<feature_2>[TAB]...[TAB]<label>


    Usage:
        extract_features -h | --help
        extract_features <features> --host=<url> [--rev-labels=<path>]
                                                 [--value-labels=<path>]
                                                 [--include-revid]
                                                 [--extractors=<num>]
                                                 [--verbose] [--debug]

    Options:
        -h --help               Print this documentation
        <features>              Classpath to a list/tuple of features
        --host=<url>            The url pointing to a MediaWiki API to use
                                for extracting features
        --rev-labels=<path>     Path to a file containing rev_id-label pairs
                                [default: <stdin>]
        --value-labels=<path>   Path to a file to write feature-labels to
                                [default: <stdout>]
        --include-revid         If set, include the revision ID as the first
                                column in the output TSV
        --extractors=<num>      The number of extractors to run in parallel
                                [default: <cpu count>]
        --verbose               Print dots and stuff
        --debug                 Print debug logging
"""
import logging
import sys
from multiprocessing import Pool, cpu_count

import docopt
import mwapi
import yamlconf

from ..errors import RevisionNotFound
from ..extractors import APIExtractor
from .util import encode

logger = logging.getLogger(__name__)


def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)

    logging.basicConfig(
        level=logging.WARNING if not args['--debug'] else logging.DEBUG,
        format='%(asctime)s %(levelname)s:%(name)s -- %(message)s'
    )

    features = yamlconf.import_module(args['<features>'])

    session = mwapi.Session(args['--host'],
                            user_agent="Revscoring feature extractor utility")
    extractor = APIExtractor(session)

    if args['--rev-labels'] == "<stdin>":
        rev_labels = read_rev_labels(sys.stdin)
    else:
        rev_labels = read_rev_labels(open(args['--rev-labels']))

    if args['--value-labels'] == "<stdout>":
        value_labels = sys.stdout
    else:
        value_labels = open(args['--value-labels'], 'w')

    include_revid = bool(args['--include-revid'])

    if args['--extractors'] == "<cpu count>":
        extractors = cpu_count()
    else:
        extractors = int(extractors)

    verbose = args['--verbose']
    debug = args['--debug']

    run(rev_labels, value_labels, features, extractor, include_revid,
        extractors, verbose, debug)


def read_rev_labels(f):
    # Check if first line is a header
    rev_id, label = f.readline().strip().split("\t")
    if rev_id != "rev_id":
        yield int(rev_id), label

    for line in f:
        rev_id, label = line.strip().split('\t')
        yield int(rev_id), label


def run(rev_labels, value_labels, features, extractor, include_revid,
        extractors, verbose, debug):
    logging.basicConfig(
        level=logging.WARNING if not debug else logging.DEBUG,
        format='%(asctime)s %(levelname)s:%(name)s -- %(message)s'
    )

    extractor_context = ConfiguredExtractor(extractor, features)
    extractor_pool = Pool(processes=extractors)

    error_rev_label_features = extractor_pool.imap(extractor_context.extract,
                                                   rev_labels)

    for error, rev_id, label, feature_values in error_rev_label_features:
        if isinstance(error, RevisionNotFound):
            if verbose:
                sys.stderr.write("?")
                sys.stderr.flush()
        elif error is not None:
            logger.error("An error occured while processing {0}:"
                         .format(rev_id))
            logger.error("\t{0}: {1}"
                         .format(error.__class__.__name__,
                                 str(error)))
        else:
            fields = list(feature_values) + [label]

            if include_revid:
                fields = [rev_id] + fields

            value_labels.write("\t".join(encode(v)
                                         for v in fields))
            value_labels.write("\n")

            if verbose:
                sys.stderr.write(".")
                sys.stderr.flush()

    if verbose:
        sys.stderr.write("\n")


class ConfiguredExtractor:

    def __init__(self, extractor, features):
        self.extractor = extractor
        self.features = features

    def extract(self, rev_label):
        rev_id, label = rev_label
        try:
            feature_values = list(self.extractor.extract(rev_id,
                                                         self.features))
            error = None
        except Exception as e:
            feature_values = None
            error = e

        return error, rev_id, label, feature_values



"""
.. autoclass:: revscoring.Feature
    :members:
"""
from math import log as math_log

from ..dependencies import Dependent

# Sets up refences to overloaded function names
math_max = max
math_min = min


class Feature(Dependent):
    """
    Represents a predictive feature.

    :Parameters:
        name : str
            The name of the feature
        process : `func`
            A function that will generate a feature value
        return_type : `type`
            A type to compare the return of this function to.
        dependencies : `list`(`hashable`)
                An ordered list of dependencies that correspond
                to the `*args` of `process`
    """
    def __init__(self, name, process=None, *, returns=None, depends_on=None):
        super().__init__(name, process, depends_on)
        self.returns = returns

    def __call__(self, *args, **kwargs):
        value = super().__call__(*args, **kwargs)

        if __debug__:
            return self.validate(value)
        else:
            return value

    def __hash__(self):
        return hash(('feature', self.name))

    # Binary math
    def __add__(self, summand):
        return add(self, summand)

    def __sub__(self, subband):
        return sub(self, subband)

    def __truediv__(self, divisor):
        return div(self, divisor)

    def __mul__(self, multiplier):
        return mul(self, multiplier)

    # Comparison
    def __lt__(self, other):
        return lt(self, other)

    def __le__(self, other):
        return le(self, other)

    def __eq__(self, other):
        return eq(self, other)

    def __ne__(self, other):
        return ne(self, other)

    def __gt__(self, other):
        return gt(self, other)

    def __ge__(self, other):
        return ge(self, other)

    # Boolean operators
    def and_(self, other):
        return and_(self, other)

    def or_(self, other):
        return or_(self, other)

    def not_(self):
        return not_(self)

    def validate(self, value):
        if isinstance(value, self.returns):
            return value
        else:
            raise ValueError("Expected {0}, but got {1} instead."
                             .format(self.returns, type(value)))

    @classmethod
    def or_constant(self, val):
        if isinstance(val, Feature):
            return val
        else:
            return Constant(val)


class Constant(Feature):

    def __init__(self, value):
        self.value = value
        super().__init__(str(value), self._process,
                         returns=type(value), depends_on=[])

    def _process(self):
        return self.value


class Modifier(Feature):
    pass


class BinaryOperator(Modifier):

    CHAR = "?"

    def __init__(self, left, right, returns=None):
        left = Feature.or_constant(left)
        right = Feature.or_constant(right)

        name = "({0} {1} {2})".format(left.name, self.CHAR, right.name)
        if returns is None:
            returns = type(self.operate(left.returns(), right.returns()))

        super().__init__(name, self.operate, returns=returns,
                         depends_on=[left, right])

    def operate(self, left, right):
        raise NotImplementedError()


class add(BinaryOperator):
    """
    Generates a feature that represents the addition of
    two :class:`revscoring.Feature` or constant values.
    """

    CHAR = "+"

    def operate(self, left, right):
        return left + right


class sub(BinaryOperator):
    """
    Generates a feature that represents the subtraction of
    two :class:`revscoring.Feature` or constant values.
    """

    CHAR = "-"

    def operate(self, left, right):
        return left - right


class mul(BinaryOperator):
    """
    Generates a feature that represents the multiplacation of
    two :class:`revscoring.Feature` or constant values.
    """

    CHAR = "*"

    def operate(self, left, right):
        return left * right


class div(BinaryOperator):
    """
    Generates a feature that represents the division of
    two :class:`revscoring.Feature` or constant values.
    """

    CHAR = "/"

    def __init__(self, left, right):
        # Explicitly setting return type to float.
        super().__init__(left, right, returns=float)

    def operate(self, left, right):
        return left / right


class Comparison(BinaryOperator):

    def __init__(self, left, right):
        # Explicitly setting return type to boolean.
        super().__init__(left, right, returns=bool)


class gt(Comparison):
    """
    Generates a feature that represents the greater-than relationship of
    two :class:`revscoring.Feature` or constant values.
    """

    CHAR = ">"

    def operate(self, left, right):
        return left > right


class lt(Comparison):
    """
    Generates a feature that represents the less-than relationship of
    two :class:`revscoring.Feature` or constant values.
    """

    CHAR = "<"

    def operate(self, left, right):
        return left < right


class ge(Comparison):
    """
    Generates a feature that represents the greater-than-or-equal relationship
    of two :class:`revscoring.Feature` or constant values.
    """

    CHAR = ">="

    def operate(self, left, right):
        return left >= right


class le(Comparison):
    """
    Generates a feature that represents the less-than-or-equal relationship of
    two :class:`revscoring.Feature` or constant values.
    """

    CHAR = "<="

    def operate(self, left, right):
        return left <= right


class eq(Comparison):
    """
    Generates a feature that represents the equality of two
    :class:`revscoring.Feature` or constant values.
    """

    CHAR = "=="

    def operate(self, left, right):
        return left == right


class ne(Comparison):
    """
    Generates a feature that represents the inequality of two
    :class:`revscoring.Feature` or constant values.
    """

    CHAR = "!="

    def operate(self, left, right):
        return left != right


class and_(Comparison):
    """
    Generates a feature that represents the conjunction of two
    :class:`revscoring.Feature` or constant values.
    """

    CHAR = "and"

    def operate(self, left, right):
        return left and right


class or_(Comparison):
    """
    Generates a feature that represents the disjunction of two
    :class:`revscoring.Feature` or constant values.
    """

    CHAR = "or"

    def operate(self, left, right):
        return left or right


class max(Modifier):
    """
    Generates a feature that represents the maximum of a set of
    :class:`revscoring.Feature` or constant values.
    """
    def __init__(self, *args):
        dependencies = [Feature.or_constant(arg) for arg in args]
        returns = float
        # Hardcoded even though max can return strings, it
        # shouldn't ever do that

        name = "max({0})".format(", ".join(f.name for f in dependencies))
        super().__init__(name, self._process, returns=returns,
                         depends_on=dependencies)

    def _process(self, *feature_values):
        return float(math_max(*feature_values))


class min(Modifier):
    """
    Generates a feature that represents the minimum of a set of
    :class:`revscoring.Feature` or constant values.
    """
    def __init__(self, *args):
        dependencies = [Feature.or_constant(arg) for arg in args]
        returns = float
        # Hardcoded even though max can return strings, it
        # shouldn't ever do that

        name = "min({0})".format(", ".join(f.name for f in dependencies))
        super().__init__(name, self._process, returns=returns,
                         depends_on=dependencies)

    def _process(self, *feature_values):
        return float(math_min(*feature_values))


class log(Modifier):
    """
    Generates a feature that represents the log of a
    :class:`revscoring.Feature`'s value.
    """
    def __init__(self, feature):
        feature = Feature.or_constant(feature)
        super().__init__("log({0})".format(feature.name), self._process,
                         returns=float, depends_on=[feature])

    def _process(self, feature_value):
        return math_log(feature_value)


class not_(Modifier):
    """
    Generates a feature that represents the negation of a
    :class:`revscoring.Feature`'s value.
    """
    def __init__(self, feature):
        feature = Feature.or_constant(feature)
        super().__init__("not {0}".format(feature.name), self._process,
                         returns=bool, depends_on=[feature])

    def _process(self, feature_value):
        return not feature_value



from nose.tools import eq_

from ...features import Feature
from ..sklearn_classifier import ScikitLearnClassifier


def test_sklean_classifier():
    skc = ScikitLearnClassifier([Feature("foo")], classifier_model=None,
                                version="0.0.1")

    eq_(skc.version, "0.0.1")



from mwtypes import Timestamp
from nose.tools import eq_

from .. import api


def test_user_info_from_doc():

    doc = {
        "userid": 24278012,
        "name": "Hoardablehotsauce",
        "editcount": 5,
        "registration": "2015-02-28T22:25:37Z",
        "groups": [
            "*",
            "user"
        ],
        "implicitgroups": [
            "*",
            "user"
        ],
        "blockid": "5752570",
        "blockedby": "Cryptic",
        "blockedbyid": "295294",
        "blockedtimestamp": "2015-02-28T22:43:23Z",
        "blockreason": "{{uw-softerblock}} <!-- Promotional username, "
                       "soft block -->",
        "blockexpiry": "infinity",
        "gender": "unknown"
    }

    info = api.APIExtractor.user_info_from_doc(doc)

    eq_(info.name, "Hoardablehotsauce")
    eq_(info.groups, ['*', "user"])
    eq_(info.implicitgroups, ['*', "user"])
    eq_(info.registration, Timestamp("2015-02-28T22:25:37Z"))
    eq_(info.block_id, 5752570)
    eq_(info.blocked_by, "Cryptic")
    eq_(info.blocked_by_id, 295294)
    eq_(info.blocked_timestamp, Timestamp("2015-02-28T22:43:23Z"))
    eq_(info.block_reason,
        "{{uw-softerblock}} <!-- Promotional username, soft block -->")
    eq_(info.block_expiry, "infinity")
    eq_(info.gender, "unknown")

    doc = {
        "userid": 24278012,
        "name": "Hoardablehotsauce",
        "editcount": 5,
        "groups": [
            "*",
            "user"
        ],
        "implicitgroups": [
            "*",
            "user"
        ],
        "gender": "unknown"
    }

    info = api.APIExtractor.user_info_from_doc(doc)
    eq_(info.registration, None)
    eq_(info.blocked_timestamp, None)

    doc = None
    info = api.APIExtractor.user_info_from_doc(doc)
    eq_(info, None)


def test_revision_metadata_from_doc():
    doc = {
        "revid": 3456789,
        "parentid": 54678,
        "comment": "Wat?",
        "user": "EpochFail",
        "userid": 34567890,
        "timestamp": "2015-01-07T12:23:57Z",
        "page": {
            "pageid": 347,
            "title": "Hats",
            "ns": 0
        }
    }

    metadata = api.APIExtractor.revision_metadata_from_doc(doc)

    eq_(metadata.rev_id, 3456789)
    eq_(metadata.parent_id, 54678)
    eq_(metadata.user_id, 34567890)
    eq_(metadata.user_text, "EpochFail")
    eq_(metadata.timestamp, Timestamp("2015-01-07T12:23:57Z"))
    eq_(metadata.comment,  "Wat?")
    eq_(metadata.page_id, 347)
    eq_(metadata.page_namespace, 0)
    eq_(metadata.page_title, "Hats")


def test_namespace_map_from_doc():
    doc = {
        "namespaces": {
            "0": {
                "id": 0,
                "case": "first-letter",
                "*": "",
                "content": ""
            },
            "1": {
                "id": 1,
                "case": "first-letter",
                "*": "Discuss\u00e3o",
                "subpages": "",
                "canonical": "Talk"
            },
            "2": {
                "id": 2,
                "case": "first-letter",
                "*": "Usu\u00e1rio(a)",
                "subpages": "",
                "canonical": "User"
            }
        },
        "namespacealiases": [
            {
                "id": 1,
                "*": "WAFFLES"
            }
        ]
    }

    namespace_map = api.APIExtractor.namespace_map_from_doc(doc)

    eq_(len(namespace_map), 3)
    eq_(sum(ns.content for ns in namespace_map.values()), 1)


def test_from_config():
    config = {
        'extractors': {
            'enwiki': {
                'host': "https://en.wikipedia.org",
                'api_path': "/w/api.php",
                'timeout': 20,
                'user_agent': "revscoring tests"
            }
        }
    }

    api.APIExtractor.from_config(config, 'enwiki')  # Doesn't error



from ...features import Feature


class Infonoise(Feature):

    def __init__(self, name, stopwords, stem_word, words_source):
        self.stopwords = set(stopwords)
        self.stem_word = stem_word
        super().__init__(name, self.process, returns=float,
                         depends_on=[words_source])

    def process(self, words):
        non_stopwords = (w for w in words if w.lower() not in self.stopwords)
        non_stopword_stems = (self.stem_word(w) for w in non_stopwords)

        length_of_stemmed = sum(len(w) for w in non_stopword_stems)

        if len(words) > 0:
            length_of_words = sum(len(w) for w in words)
        else:
            length_of_words = 0

        return length_of_stemmed / max(length_of_words, 1)



from collections import namedtuple

from mwtypes import Timestamp
from nose.tools import eq_

from ...datasources import previous_user_revision, revision
from ...dependencies import solve
from ..previous_user_revision import seconds_since


def test_seconds_since():
    FakeRevisionMetadata = namedtuple("FakeRevisionMetadata",
                                      ['timestamp'])

    cache = {
        revision.metadata: FakeRevisionMetadata(Timestamp(10)),
        previous_user_revision.metadata: FakeRevisionMetadata(Timestamp(1))
    }
    eq_(solve(seconds_since, cache=cache), 9)

    # Makes sure we don't crash when there was no previous user revision
    cache = {
        revision.metadata: FakeRevisionMetadata(Timestamp(10)),
        previous_user_revision.metadata: None
    }
    eq_(solve(seconds_since, cache=cache), 0)



"""
A collection of Naive Bayes type classifier models.

.. autoclass:: revscoring.scorer_models.GaussianNB
    :members:
    :member-order:

.. autoclass:: revscoring.scorer_models.MultinomialNB
    :members:
    :member-order:

.. autoclass:: revscoring.scorer_models.BernoulliNB
    :members:
    :member-order:
"""
import logging

from sklearn import naive_bayes

from .sklearn_classifier import ScikitLearnClassifier

logger = logging.getLogger("revscoring.scorers.nb")


class NB(ScikitLearnClassifier):
    def __init__(self, features, *, version=None, nb=None,
                 sklearn_class=None, **kwargs):
        if nb is None:
            nb = sklearn_class(**kwargs)
        super().__init__(features, classifier_model=nb, version=version)
NBModel = NB
"Alias for backwards compatibility"


class GaussianNB(NBModel):
    """
    Implements a Gaussian Naive Bayes model.

    :Params:
        features : `list` ( :class:`revscoring.Feature` )
            The features that the model will be trained on
        version : str
            A version string representing the version of the model
        `**kwargs`
            Passed to :class:`sklearn.naive_bayes.GaussianNB`
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, sklearn_class=naive_bayes.GaussianNB, **kwargs)
GaussianNBModel = GaussianNB
"Alias for backwards compatibility"


class MultinomialNB(NBModel):
    """
    Implements a Multinomial Naive Bayes model.

    :Params:
        features : `list` ( :class:`revscoring.Feature` )
            The features that the model will be trained on
        version : str
            A version string representing the version of the model
        `**kwargs`
            Passed to :class:`sklearn.naive_bayes.MultinomialNB`
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, sklearn_class=naive_bayes.MultinomialNB,
                         **kwargs)
MultinomialNBModel = MultinomialNB
"Alias for backwards compatibility"


class BernoulliNB(NBModel):
    """
    Implements a Bernoulli Naive Bayes model.

    :Params:
        features : `list` ( :class:`revscoring.Feature` )
            The features that the model will be trained on
        version : str
            A version string representing the version of the model
        `**kwargs`
            Passed to :class:`sklearn.naive_bayes.BernoulliNB`
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, sklearn_class=naive_bayes.BernoulliNB,
                         **kwargs)
BernoulliNBModel = BernoulliNB
"Alias for backwards compatibility"



"""
A collection of Random Forest type classifier models.

.. autoclass:: revscoring.scorer_models.RF
    :members:
    :member-order:
"""
import logging

from sklearn.ensemble import RandomForestClassifier

from .sklearn_classifier import ScikitLearnClassifier

logger = logging.getLogger("revscoring.scorers.rf")


class RF(ScikitLearnClassifier):
    """
    Implements a Random Forest model.

    :Params:
        features : `list` ( :class:`revscoring.Feature` )
            The features that the model will be trained on
        version : str
            A version string representing the version of the model
        `**kwargs`
            Passed to :class:`sklearn.ensemble.RandomForestClassifier`
    """
    def __init__(self, features, *, version=None, rf=None,
                 **kwargs):

        if rf is None:
            rf = RandomForestClassifier(**kwargs)

        super().__init__(features, classifier_model=rf, version=version)
RFModel = RF
"Alias for backwards compatibility"



import json
import random
from io import BytesIO
from itertools import chain

from nose.tools import eq_

from ...features import Feature
from ..scorer_model import MLScorerModel


def process_float():
    return float()
some_float = Feature("some_float", process_float(),
                     depends_on=[], returns=float)


def process_other_float():
    return float()
other_float = Feature("other_float", process_other_float(),
                      depends_on=[], returns=float)
FEATURES = [some_float, other_float]


def train_score(model):
    deterministic = random.Random(0)
    observations = list(chain(
        zip(((some, other) for some, other in
             zip((deterministic.normalvariate(1, .3) for i in range(500)),
                 (deterministic.normalvariate(2, .5) for i in range(500)))),
            (False for i in range(500))),
        zip(((some, other) for some, other in
             zip((deterministic.normalvariate(-1, .5) for i in range(50)),
                 (deterministic.normalvariate(-2, .3) for i in range(50)))),
            (True for i in range(50)))
    ))
    deterministic.shuffle(observations)

    mid = int(len(observations)/2)
    train_set = observations[:mid]
    test_set = observations[mid:]

    model.train(train_set)
    score_doc = model.score((-1, -2))

    eq_(score_doc['prediction'], True)
    assert score_doc['probability'][True] > 0.5, \
        "Probability of True {0} is not > 0.5" \
        .format(score_doc['probability'][True])
    json.dumps(score_doc)  # Checks if the doc is JSONable

    test_stats = model.test(test_set)

    assert test_stats['roc']['auc'] > 0.5


def pickle_and_unpickle(model):
    f = BytesIO()
    model.dump(f)
    f.seek(0)  # Rewind the file
    reconstructed_model = MLScorerModel.load(f)
    eq_([feature.name for feature in reconstructed_model.features],
        [feature.name for feature in model.features])
    eq_(type(reconstructed_model), type(model))
    train_score(reconstructed_model)


def get_and_format_info(model):

    assert model.info() is not None
    print(model.format_info())



import pickle

from nose.tools import eq_

from .. import english
from ...datasources import revision
from ...dependencies import solve
from .util import compare_extraction

BAD = [
    "ass", "arse", "ASS", "assface", "asses", "stupidass", "fatass", "lazyass",
    "assclown", "asshat", "stupidasshat", "stupidarsehat",
    "autofellate", "autofellation", "autofellatio",
    "bitch", "biotch", "BITCH", "bitchface", "bitches", "supidbitch",
    "asdasdbitchasdlnasla",
    "blowjob", "blowme",
    "bollocks",
    "booger", "boogers", "boogereater", "boooooger"
    "bootlip",
    "bugger",
    "butt", "buttfuck", "buttclown", "assbuttclown",
    "clunge",
    "cock", "cocks", "cocking", "cockface",
    "coon", "coooon",
    "cracker", "kracker",
    "crackhead",
    "crook",
    "cunt", "CUNT", "cuntface", "cunts", "stupidcunt",
    "dick", "dicks", "dicking", "dickface", "limpdick",
    "dildo",
    "dishonest",
    "defraud",
    "dothead",
    "dyke",
    "fag", "faggot", "phaggot", "faaaag", "fagface", "fags",
    "fart", "farteater", "farter", "farting",
    "fuck", "FUCK", "fuckface", "fucker", "stupidfuck", "whatthefuck",
    "gay", "gaaaay", "gays", "ghey",
    "gyp", "gypie", "gippo", "gyppie",
    "gook", "goook",
    "hillbilly", "hill-billy",
    "hooker",
    "homosexual", "homo",
    "injun",
    "jap", "jappo",
    "kike", "kyke", "kiike",
    "kwashee", "kwashi",
    "lesbian",
    "liar",
    "meth", "methhead",
    "nazi", "nazzzi",
    "nigger", "nigga", "nig", "niglet", "nigra",
    "nonce", "noncer", "noncing",
    "overdosed",
    "peckerwood",
    "penis", "peni", "penises",
    "pedo", "pedophile", "paedophyles", "pedofile",
    "piss", "pissface", "pisser",
    "pothead",
    "prostitute", "prostituteface"
    "qwashi", "qwashee",
    "raghead", "rag-head",
    "redneck", "red-neck",
    "redskin", "red-skin",
    "roundeye", "round-eye",
    "satanic", "satanists",
    "shit", "SHIT", "shitface", "shitting", "stupidshit", "shite", "shat",
    "slut", "SLUT", "slutface", "slutty",
    "spik", "spick", "spig", "spic",
    "subnormal",
    "squarehead",
    "terrorist", "terrorising", "terrorised", "terrorized", "terrorists",
    "theif", "theives",
    "tits", "titties", "tities", "titty",
    "transexual", "tranny",
    "twat", "TWAT", "twatface", "twatting", "stupidtwat",
    "vagina", "vajaja", "vag",
    "wank", "wanker", "wanka", "wankface",
    "wetback", "wetbacker",
    "whore", "WHORE", "whoreface", "whoring", "stupidwhore",
    "wog", "woggg",
    "wop", "woppp",
    "yid",
    "zipperhead"
]

INFORMAL = [
    "ain't", "aint",
    "awesome", "AWESOME", "aaawsome", "awesom",
    "bla", "blah", "blaaahhh", "blahblahblahblaaaaa",
    "bro",
    "bye", "byebye",
    "boner", "booner",
    "can't", "cant",
    "cool", "kewl", "cooler", "coolest", "coolass",
    "chug",
    "crap", "crappp", "crappiest", "crappier",
    "don't", "dont",
    "dude", "duuuuude",
    "dumb", "dummy", "dumbest", "dummies",
    "dad", "daddy", "dada",
    "goodbye", "good-bye",
    "hi", "hihi", "ha", "haha", "hehe", "ho", "hoho", "hu", "huhu", "muhaha",
    "mwuhahaha",
    "hello", "helo", "hellloooo",
    "hey", "heeeey", "haay",
    "hm", "hmmmm", "hhhmmmm",
    "i", "I",
    "idiot", "idiooot",
    "lalala", "lalalala",
    "lol", "lololol", "LLLolLLLlOL",
    "loser",
    "love", "looove", "luv",
    "meow", "meeeoooow",
    "mom", "mommy", "momma",
    "moron",
    "munch",
    "nope",
    "nerd", "nerds",
    "noob", "noobs",
    "omg",
    "ok", "okay",
    "poop", "poooops"
    "pretty",
    "retard", "retarded", "tard",
    "rofl",
    "sexy", "sexxxy",
    "smelly",
    "soo", "soooooo",
    "sorry",
    "stupid", "stooopid", "stupidface", "brainstupid"
    "stink", "stinky", "stinks",
    "suck", "sucker", "sucking", "sux",
    "shouldn't", "shouldnt",
    "test edit",
    "turd", "turds",
    "wasn't", "wasnt",
    "whatsup", "wuzzzap", "wussup",
    "wuz",
    "won't", "wont",
    "woof",
    "wtf", "omgwtfbbq",
    "ya'll", "yall",
    "yay", "yyyaaaayyyy",
    "yeah", "yea", "yeaaaa",
    "you", "you've", "you're", "you'll",
    "yolo", "yooolloo"
]

OTHER = [
    """
    The Smyth Report is the common name of an administrative history written
    by physicist Henry DeWolf Smyth about the Manhattan Project, the Allied
    effort to develop atomic bombs during World War II. It was released to the
    public on August 12, 1945, just days after the atomic bombings of Hiroshima
    and Nagasaki. Smyth was commissioned to write the report by Major General
    Leslie Groves, the director of the Manhattan Project. The Smyth Report was
    the first official account of the development of the atomic bombs and the
    basic physical processes behind them. Since anything in the declassified
    Smyth Report could be discussed openly, it focused heavily on basic nuclear
    physics and other information which was either already widely known in the
    scientific community or easily deducible by a competent scientist. It
    omitted details about chemistry, metallurgy, and ordnance, ultimately
    giving a false impression that the Manhattan Project was all about physics.
    """,
    'association', 'shihtzhu', 'Wafflehats', "he", "hay", "morass", "wood",
    "pecker", 'suction', 'vaginal', 'titillatingly', 'test', 'edit'
]


def test_badwords():
    compare_extraction(english.revision.badwords_list, BAD, OTHER)


def test_informals():
    compare_extraction(english.revision.informals_list, INFORMAL, OTHER)


def test_revision():
    # Words
    cache = {revision.text: "I have an m80; and a shovel."}
    eq_(solve(english.revision.words_list, cache=cache),
        ["I", "have", "an", "m80", "and", "a", "shovel"])

    # Misspellings
    cache = {revision.text: 'This is spelled worngly. <td>'}
    eq_(solve(english.revision.misspellings_list, cache=cache), ["worngly"])

    # Infonoise
    cache = {revision.text: "This is running!"}
    eq_(solve(english.revision.infonoise, cache=cache), 3/13)


def test_pickling():

    eq_(english, pickle.loads(pickle.dumps(english)))



from .feature import Constant, Feature, Modifier


def trim(features, context=None):
    context = context or {}
    cache = set()

    if hasattr(features, "__iter__"):
        for feature in features:
            yield from _trim(feature, context, cache)
    else:
        yield from _trim(features, context, cache)


def _trim(dependent, context, cache):
    if isinstance(dependent, Feature):
        feature = dependent
        if isinstance(feature, Modifier):
            for dependent in feature.dependencies:
                yield from _trim(dependent, context, cache)
        elif isinstance(feature, Constant):
            pass
        else:
            cache.add(feature)
            yield feature



import pickle
from collections import namedtuple

import enchant
from nose.tools import eq_

from ....datasources import parent_revision, revision
from ....dependencies import solve
from ..space_delimited import SpaceDelimited


def test_no_params():
    SpaceDelimited("fake")  # Should not error


def test_words():
    sd = SpaceDelimited("fake")

    cache = {revision.text: "Foobar bad badwords hat 75 m80, <td> {{foo}}"}
    eq_(solve(sd.revision.words_list, cache=cache),
        ["Foobar", "bad", "badwords", "hat", "m80", "foo"])
    eq_(solve(sd.revision.content_words_list, cache=cache),
        ["Foobar", "bad", "badwords", "hat", "m80"])

    cache = {parent_revision.text: None}
    eq_(solve(sd.parent_revision.words_list, cache=cache), [])
    eq_(solve(sd.parent_revision.content_words_list, cache=cache), [])


def test_badwords():
    sd = SpaceDelimited("fake", badwords=[r"bad(words)?"])

    cache = {revision.text: "Foobar bad badwords hat."}
    eq_(solve(sd.revision.badwords_list, cache=cache), ["bad", "badwords"])

    assert hasattr(sd.revision, "badwords")
    assert hasattr(sd.parent_revision, "badwords")
    assert hasattr(sd.diff, "badwords_added")
    assert hasattr(sd.diff, "badwords_removed")

    cache = {parent_revision.text: None}
    eq_(solve(sd.parent_revision.badwords_list, cache=cache), [])


def test_informals():
    sd = SpaceDelimited("fake", informals=[r"inform(als)?"])

    cache = {revision.text: "Foobar inform informals als hat."}
    eq_(solve(sd.revision.informals_list, cache=cache),
        ["inform", "informals"])

    assert hasattr(sd.revision, "informals")
    assert hasattr(sd.parent_revision, "informals")
    assert hasattr(sd.diff, "informals_added")
    assert hasattr(sd.diff, "informals_removed")

    cache = {parent_revision.text: None}
    eq_(solve(sd.parent_revision.informals_list, cache=cache), [])


def test_infonoise():
    Stemmer = namedtuple("Stemmer", ["stem"])
    stemmer = Stemmer(lambda w: w[0])  # First char
    stopwords = set(["stop", "word"])
    sd = SpaceDelimited("fake", stemmer=stemmer, stopwords=stopwords)

    cache = {revision.text: "Waffle stop word"}
    eq_(solve(sd.revision.infonoise, cache=cache), 1/len("Wafflestopword"))

    cache = {parent_revision.text: "Waffle stop word"}
    eq_(solve(sd.parent_revision.infonoise, cache=cache),
        1/len("Wafflestopword"))

    assert hasattr(sd.revision, "infonoise")
    assert hasattr(sd.parent_revision, "infonoise")


def test_misspellings():
    Dictionary = namedtuple("Dictionary", ["check"])
    dictionary = Dictionary(lambda w: w != "misspelled")

    sd = SpaceDelimited("fake", dictionary=dictionary)

    cache = {revision.text: "Foobar misspelled als hat."}
    eq_(solve(sd.revision.misspellings_list, cache=cache), ["misspelled"])

    assert hasattr(sd.revision, "misspellings")
    assert hasattr(sd.parent_revision, "misspellings")
    assert hasattr(sd.diff, "misspellings_added")
    assert hasattr(sd.diff, "misspellings_removed")

    cache = {parent_revision.text: None}
    eq_(solve(sd.parent_revision.misspellings_list, cache=cache), [])


def test_utf16_issue():
    dictionary = enchant.Dict('en')

    sd = SpaceDelimited("fake", dictionary=dictionary)

    cache = {revision.text: "𐎤𐎢𐎽𐎢𐏁"}
    eq_(solve(sd.revision.misspellings_list, cache=cache), ["𐎤𐎢𐎽𐎢𐏁"])


BADWORDS = [r"bad(words)?"]
Dictionary = namedtuple("Dictionary", ["check"])


def check(word):
    return word != "misspelled"

DICTIONARY = Dictionary(check)
INFORMALS = [r"inform(als)?"]
Stemmer = namedtuple("Stemmer", ["stem"])


def stem(word):
    return word[0]  # First char

STEMMER = Stemmer(stem)
STOPWORDS = set(["stop", "word"])


def test_pickle():
    sd = SpaceDelimited("fake", badwords=BADWORDS, dictionary=DICTIONARY,
                        informals=INFORMALS, stemmer=STEMMER,
                        stopwords=STOPWORDS)

    pickled_sd = pickle.dumps(sd)
    print(pickled_sd)
    unpickled_sd = pickle.loads(pickled_sd)
    eq_(unpickled_sd, sd)



import pickle

from nose.tools import eq_

from .. import estonian
from ...datasources import revision
from ...dependencies import solve
from .util import compare_extraction

BAD = [
    "homo", "homod", "jobu", "junn", "junni", "kaka", "lits", "loll",
    "lollakas", "lollid", "munn", "munni", "munnid", "neeger", "nigga",
    "noku", "pask", "pede", "peded", "pedekas", "pederast", "pederastid",
    "perse", "perses", "persse", "pihku", "putsi", "sitane", "sitt", "sitta",
    "tsmir", "türa", "tšmir", "vittu", "vitupea"
]

INFORMAL = [
    "haha", "hahaha", "jou", "noob", "raisk", "räme", "sakib", "suht",
    "tegelt", "tere", "tsau"
]

OTHER = [
    """
    Friedrich Wilhelm Rembert von Berg sündis Beļava ja Sangaste mõisniku
    Friedrich Georg von Bergi ja Gerdruta Wilhelmine von Ermesi vanima pojana.
    Tal olid nooremad vennad Gustav, Magnus ja Alexander. Friedrich Wilhelmi ja
    tema vendade koduõpetaja oli hilisem tuntud astronoom Wilhelm Struve.
    """
]


def test_badwords():
    compare_extraction(estonian.revision.badwords_list, BAD, OTHER)


def test_informals():
    compare_extraction(estonian.revision.informals_list, INFORMAL, OTHER)


def test_revision():
    # Words
    cache = {revision.text: "Tal olid nooremad m80, vennad Gustav."}
    eq_(solve(estonian.revision.words_list, cache=cache),
        ["Tal", "olid", "nooremad", "m80", "vennad", "Gustav"])

    # Misspellings
    cache = {revision.text: "Tal olid nooremad, vennad, worngly. <td>"}
    eq_(solve(estonian.revision.misspellings_list, cache=cache), ["worngly"])


def test_pickling():

    eq_(estonian, pickle.loads(pickle.dumps(estonian)))



from ..rf import RFModel
from .util import (FEATURES, get_and_format_info, pickle_and_unpickle,
                   train_score)


def test_rf():
    model = RFModel(FEATURES)
    get_and_format_info(model)
    train_score(model)
    pickle_and_unpickle(model)
    get_and_format_info(model)



import pickle

from nose.tools import eq_

from .. import dutch
from ...datasources import revision
from ...dependencies import solve
from .util import compare_extraction

BAD = [
    "aars",
    "anaal", "anus",
    "balhaar",
    "drol", "drollen",
    "fack", "facking",
    "flikker", "flikkers",
    "focking",
    "geil", "geile", "gelul",
    "geneukt",
    "hoer", "hoere", "hoeren",
    "homo", "homos",
    "kak", "kaka",
    "kakhoofd", "kakken",
    "kanker", "kenker",
    "klootzak", "klootzakken",
    "klote",
    "kont", "kontgat",
    "kontje",
    "pedo",
    "penis", "penissen",
    "peop",
    "piemel", "piemels",
    "pijpen",
    "pik",
    "pimel",
    "pipi",
    "poep", "poepchinees", "poepen", "poephoofd",
    "poepie", "poepje", "poepjes", "poepsex", "poept", "poepte",
    "porn", "porno",
    "neuk", "neuke",
    "neuken", "neukende",
    "neukt", "neukte", "neukten",
    "strond", "stront",
    "suck", "sucks", "suckt",
    "zuigt",
    "sukkel", "sukkels",
    "tering", "tetten", "tieten",
    "vagina",
    "verekte", "verkracht",
    "dikzak",
    "dildo",
    "mogolen", "mogool", "mongool", "mooiboy",
    "neger", "negers",
    "shit",
    "sperma",
    "kut", "kutje", "kutjes",
    "stelletje",
    "loser", "losers",
    "lul", "lullen",
    "reet",
    "scheet", "scheten",
    "schijt",
    "diaree",
    "slet",
    "lekkerding",
    "likken"
]

INFORMAL = [
    "aap", "aapjes",
    "banaan",
    "bent",
    "boe", "boeit",
    "doei"
    "dombo", "domme",
    "eigelijk",
    "godverdomme",
    "groetjes",
    "gwn",
    "hoi",
    "hallo", "halloo",
    "heb",
    "heej", "heey", "heeel",
    "hou", "houd",
    "hoihoi", "hoii", "hoiii",
    "hoor",
    "izan",
    "jij",
    "jou",
    "jullie",
    "kaas",
    "klopt",
    "kots",
    "kusjes",
    "lekker", "lekkere", "lkkr",
    "maarja",
    "mama",
    "nou",
    "oma",
    "ofzo",
    "oke",
    "sex", "sexy",
    "snap",
    "stinken", "stinkt",
    "stoer",
    "swag",
    "swek",
    "vies", "vieze",
    "vind",
    "vuile",
    "xxx",
    "zielig",
    "zooi",
    "yolo",
    "zeg"
]

OTHER = [
    """
    De stemtoonhoogte is de toonhoogte van de kamertoon. Door middel van een
    stemvork is deze kamertoon beschikbaar voor het stemmen van een
    muziekinstrument.

    Internationaal is deze toonhoogte in het midden van de 20e eeuw vastgesteld
    op een frequentie van 440 Hz. De stemtoon lag echter niet altijd vast. Soms
    leest men ergens dat de stemtoon door de eeuwen heen steeds hoger is komen
    te liggen, maar dat is slechts de helft van het verhaal. Er waren orgels
    die een hogere stemtoon hadden, en later lager gestemd werden. Kerkorgels
    verschilden enorm van stemtoon. In de loop van de tijd is die variatie
    steeds kleiner geworden. Naarmate mensen steeds mobieler werden, ontstond
    ook de behoefte aan meer compatibiliteit van instrumenten.
    """
]


def test_badwords():
    compare_extraction(dutch.revision.badwords_list, BAD, OTHER)


def test_informals():
    compare_extraction(dutch.revision.informals_list, INFORMAL, OTHER)


def test_revision():
    # Words
    cache = {revision.text: "De stemtoonhoogte is m80 van de kamertoon."}
    eq_(solve(dutch.revision.words_list, cache=cache),
        ["De", "stemtoonhoogte", "is", "m80", "van", "de", "kamertoon"])

    # Misspellings
    cache = {revision.text: 'Door middel van een worngly. <td>'}
    eq_(solve(dutch.revision.misspellings_list, cache=cache), ["worngly"])

    # Infonoise
    cache = {revision.text: "Door middel van een!"}
    eq_(solve(dutch.revision.infonoise, cache=cache), 0.375)


def test_pickling():

    eq_(dutch, pickle.loads(pickle.dumps(dutch)))



import random


def encode(val, none_val="NULL"):
    if val is None:
        return none_val
    elif isinstance(val, bytes):
        val = str(val, 'utf-8', "replace")
    else:
        val = str(val)

    return val.replace("\t", "\\t").replace("\n", "\\n")


DECODERS = {
    'int': lambda v: int(v),
    'float': lambda v: float(v),
    'str': lambda v: str(v),
    'bool': lambda v: v in ("True", "true", "1", "T", "y", "Y")
}


def read_observations(f, features, decode_label):
    for line in f:
        parts = line.strip().split("\t")
        values = parts[:-1]
        label = parts[-1]

        label = decode_label(label)

        feature_values = []
        for feature, value in zip(features, values):

            if feature.returns == bool:
                feature_values.append(value == "True")
            else:
                feature_values.append(feature.returns(value))

        yield feature_values, label


def train_test_split(observations, test_prop=0.25):
    # Split train and test set from obs.
    observations = list(observations)
    random.shuffle(observations)

    test_set_size = int(len(observations) * test_prop)

    test_set = observations[:test_set_size]
    train_set = observations[test_set_size:]

    return train_set, test_set



"""
``revscoring train_test -h``
::

    Trains and tests a scorer model.  This utility expects to get a file of
    tab-separated feature values and labels from which to construct a model.

    Usage:
        train_test -h | --help
        train_test <scorer_model> <features> [-p=<kv>]...
                   [--version=<vers>]
                   [--values-labels=<path>]
                   [--model-file=<path>]
                   [--label-type=<type>]
                   [--test-prop=<prop>]
                   [--debug]

    Options:
        -h --help               Prints this documentation
        <scorer_model>          Classpath to an the MLScorerModel to construct
                                and train
        <features>              Classpath to an list of features to use when
                                constructing the model
        -p --parameter=<kv>     A key-value argument pair to use when
                                constructing the scorer_model.
        --version=<vers>        A version to associate with the model
        --values-labels=<path>  Path to a file containing feature values and
                                labels [default: <stdin>]
        --model-file=<math>     Path to write a model file to
                                [default: <stdout>]
        --label-type=<type>     Interprets the labels as the appropriate type
                                (int, float, str, bool) [default: str]
        --test-prop=<prop>      The proportion of data that should be withheld
                                for testing the model. [default: 0.20]
        --debug                 Print debug logging.
"""
import json
import logging
import sys

import docopt
import yamlconf

from . import util

logger = logging.getLogger(__name__)


def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)

    logging.basicConfig(
        level=logging.INFO if not args['--debug'] else logging.DEBUG,
        format='%(asctime)s %(levelname)s:%(name)s -- %(message)s'
    )

    ScorerModel = yamlconf.import_module(args['<scorer_model>'])
    features = yamlconf.import_module(args['<features>'])

    version = args['--version']

    model_kwargs = {}
    for parameter in args['--parameter']:
        key, value = parameter.split("=")
        model_kwargs[key] = json.loads(value)

    scorer_model = ScorerModel(features, version=version, **model_kwargs)

    if args['--values-labels'] == "<stdin>":
        observations_f = sys.stdin
    else:
        observations_f = open(args['--values-labels'], 'r')

    if args['--model-file'] == "<stdout>":
        model_file = sys.stdout.buffer
    else:
        model_file = open(args['--model-file'], 'wb')

    decode_label = util.DECODERS[args['--label-type']]

    observations = util.read_observations(observations_f,
                                          scorer_model.features,
                                          decode_label)

    test_prop = float(args['--test-prop'])

    run(observations, model_file, scorer_model, test_prop)


def run(observations, model_file, scorer_model, test_prop):

    scorer_model = _train_test(scorer_model, observations, test_prop)

    sys.stderr.write(scorer_model.format_info())

    sys.stderr.write("\n\n")

    scorer_model.dump(model_file)


def _train_test(scorer_model, observations, test_prop):
    train_set, test_set = util.train_test_split(observations,
                                                test_prop=test_prop)

    logger.debug("Test set: {0}".format(len(test_set)))
    logger.debug("Train set: {0}".format(len(train_set)))

    logger.info("Training model...")
    scorer_model.train(train_set)

    logger.info("Testing model...")
    scorer_model.test(test_set)

    return scorer_model



"""
.. autoclass:: Context
    :members:
"""
from .functions import dig, draw, expand, solve


class Context:
    """
    Represents a contextual space for solving dependencies

    :Parameters:
        context : dict | iterable
            A set of dependents to be used in place of those already
            provided when solving dependencies.
        cache : dict
            A cache of computed values to use for every call to
            :func:`revscoring.dependencies.solve`
    """
    def __init__(self, context=None, cache=None):
        self.cache = cache or {}

        # Make sure context is a dict
        if context is None:
            self.context = {}
        elif not isinstance(context, dict):
            self.context = {d: d for d in context}
        else:  # else leave context alone
            self.context = context

    def solve(self, dependents, context=None, cache=None):
        """
        Solves an iterable of dependents within the context.

        See :func:`~revscoring.dependencies.solve` for call
        signature.
        """
        context, cache = self.update_context_and_cache(context, cache)
        return solve(dependents, context=context, cache=cache)

    def expand(self, dependents, cache=None, context=None):
        """
        Expands iterable of all dependents within the context.

        See :func:`~revscoring.dependencies.expand` for call
        signature.
        """
        context, cache = self.update_context_and_cache(context, cache)
        return expand(dependents, context=context, cache=cache)

    def dig(self, dependents, cache=None, context=None):
        """
        Digs up the root dependents within the context.

        See :func:`~revscoring.dependencies.dig` for call signature.
        """
        context, cache = self.update_context_and_cache(context, cache)
        return dig(dependents, context=context, cache=cache)

    def draw(self, dependent, cache=None, context=None):
        """
        Returns a string representing the tree structure of a dependent's
        dependencies.

        See :func:`~revscoring.dependencies.draw` for call signature.
        """
        context, cache = self.update_context_and_cache(context, cache)
        return draw(dependent, context=context, cache=cache)

    def update_context_and_cache(self, context, cache):
        new_context = {}  # Prepare context
        new_context.update(self.context)  # Load extractor context
        new_context.update(context or {})  # Load call context

        new_cache = {}  # Prepare cache
        new_cache.update(self.cache)  # Load cache for extractor
        new_cache.update(cache or {})  # Load call cache

        return new_context, new_cache



from collections import namedtuple

from mwtypes import Timestamp
from nose.tools import eq_

from ...datasources import parent_revision, revision
from ...dependencies import solve
from ..parent_revision import (bytes, chars, markup_chars, numeric_chars,
                               seconds_since, symbolic_chars, uppercase_chars,
                               was_same_user)


def test_was_same_user():
    FakeRevisionMetadata = namedtuple("FakeRevisionMetadata",
                                      ['user_id', 'user_text'])

    cache = {
        revision.metadata: FakeRevisionMetadata(10, "Foobar"),
        parent_revision.metadata: FakeRevisionMetadata(10, "Foobar")
    }
    eq_(solve(was_same_user, cache=cache), True)

    cache = {
        revision.metadata: FakeRevisionMetadata(10, "Foobar"),
        parent_revision.metadata: FakeRevisionMetadata(10, "Fleebar")
    }
    eq_(solve(was_same_user, cache=cache), True)

    cache = {
        revision.metadata: FakeRevisionMetadata(None, "127.4.5.6"),
        parent_revision.metadata: FakeRevisionMetadata(None, "127.4.5.6")
    }
    eq_(solve(was_same_user, cache=cache), True)

    # Make sure we don't error when there is no parent revision
    cache = {
        revision.metadata: FakeRevisionMetadata(None, "127.4.5.6"),
        parent_revision.metadata: None
    }
    eq_(solve(was_same_user, cache=cache), False)


def test_seconds_since():
    FakeRevisionMetadata = namedtuple("FakeRevisionMetadata",
                                      ['timestamp'])

    cache = {
        revision.metadata: FakeRevisionMetadata(Timestamp(10)),
        parent_revision.metadata: FakeRevisionMetadata(Timestamp(1))
    }
    eq_(solve(seconds_since, cache=cache), 9)

    # Make sure we don't error when there is no parent revision
    cache = {
        revision.metadata: FakeRevisionMetadata(Timestamp(10)),
        parent_revision.metadata: None
    }
    eq_(solve(seconds_since, cache=cache), 0)


def test_bytes():
    FakeRevisionMetadata = namedtuple("FakeRevisionMetadata",
                                      ['bytes'])

    cache = {
        parent_revision.metadata: FakeRevisionMetadata(25)
    }
    eq_(solve(bytes, cache=cache), 25)

    # Make sure we don't error when there is no parent revision
    cache = {
        parent_revision.metadata: None
    }
    eq_(solve(bytes, cache=cache), 0)


def test_chars():
    cache = {
        parent_revision.text: "Twelve chars"
    }
    eq_(solve(chars, cache=cache), 12)

    # Make sure we don't error when there is no parent revision
    cache = {
        parent_revision.text: None
    }
    eq_(solve(chars, cache=cache), 0)


def test_markup_chars():
    cache = {
        parent_revision.text: "Twelve {{chars}}"
    }
    eq_(solve(markup_chars, cache=cache), 4)

    # Make sure we don't error when there is no parent revision
    cache = {
        parent_revision.text: None
    }
    eq_(solve(markup_chars, cache=cache), 0)


def test_numeric_chars():
    cache = {
        parent_revision.text: "Twelve hats pants 95 bananas!"
    }
    eq_(solve(numeric_chars, cache=cache), 2)

    # Make sure we don't error when there is no parent revision
    cache = {
        parent_revision.text: None
    }
    eq_(solve(numeric_chars, cache=cache), 0)


def test_symbolic_chars():
    cache = {
        parent_revision.text: "Twelve hats?  Pants, #95 bananas!"
    }
    eq_(solve(symbolic_chars, cache=cache), 4)

    # Make sure we don't error when there is no parent revision
    cache = {
        parent_revision.text: None
    }
    eq_(solve(symbolic_chars, cache=cache), 0)


def test_uppercase_chars():
    cache = {
        parent_revision.text: "Twelve hats?  Pants, #95 bananas!"
    }
    eq_(solve(uppercase_chars, cache=cache), 2)

    # Make sure we don't error when there is no parent revision
    cache = {
        parent_revision.text: None
    }
    eq_(solve(uppercase_chars, cache=cache), 0)



"""
This module contains a collection of utilities for extracting
:class:`~revscoring.Feature` and
:class:`~revscoring.Datasource` for a revision.

api
+++
.. automodule:: revscoring.extractors.api

extractor
+++++++++
.. automodule:: revscoring.extractors.extractor

"""
from .extractor import Extractor
from .api import APIExtractor

__all__ = [Extractor, APIExtractor]



from nose.tools import eq_, raises

from ...errors import DependencyError, DependencyLoop
from ..dependent import Dependent
from ..functions import dig, draw, expand, solve


def test_solve():
    # Simple functions
    eq_(solve(lambda: "foo"), "foo")
    eq_(list(solve([lambda: "foo", lambda: "bar"])), ["foo", "bar"])

    # Dependents
    foo = Dependent("foo", lambda: "foo")
    bar = Dependent("bar", lambda: "bar")
    foobar = Dependent("foobar", lambda foo, bar: foo + bar,
                       depends_on=[foo, bar])
    eq_(solve(foobar), "foobar")
    eq_(list(solve([foo, bar, foobar])), ["foo", "bar", "foobar"])

    # Cache
    eq_(solve(foobar, cache={bar: "baz"}), "foobaz")

    # Context
    mybar = Dependent("bar", lambda: "baz")
    eq_(solve(foobar, context={mybar}), "foobaz")
    eq_(solve(foobar, context={mybar: mybar}), "foobaz")
    eq_(solve(foobar, context={bar: mybar}), "foobaz")
    eq_(solve(foobar, context={bar: lambda: "baz"}), "foobaz")


@raises(RuntimeError)
def test_unsolveable():
    solve(5)


@raises(DependencyLoop)
def test_dependency_loop():
    foo = Dependent("foo")
    bar = Dependent("bar", depends_on=[foo])
    my_foo = Dependent("foo", depends_on=[bar])

    solve(bar, context={my_foo})


def test_expand():
    foo = Dependent("foo", lambda: "foo")
    bar = Dependent("bar", lambda: "bar")
    foobar = Dependent("foobar", lambda foo, bar: foo + bar,
                       depends_on=[foo, bar])

    derp = Dependent("derp", lambda: "derp")
    fooderp = Dependent("fooderp", lambda foo, derp: foo + derp,
                        depends_on=[foo, derp])

    dependents = list(expand(foobar))
    eq_(len(dependents), 3)
    eq_(set(dependents), {foo, bar, foobar})

    dependents = list(expand([fooderp, foobar]))
    eq_(len(dependents), 5)
    eq_(set(dependents), {derp, fooderp, foo, bar, foobar})


def test_dig():
    foo = Dependent("foo", lambda: "foo")
    bar = Dependent("bar", lambda: "bar")
    foobar = Dependent("foobar", lambda foo, bar: foo + bar,
                       depends_on=[foo, bar])
    foobar_foobar = Dependent("foobar_foobar",
                              lambda foobar1, foobar2: foobar1 + "_" + foobar2,
                              depends_on=[foobar, foobar])

    roots = list(dig(foobar_foobar))
    eq_(len(roots), 2)
    eq_(set(roots), {foo, bar})

    roots = list(dig([foobar, foo, bar]))
    eq_(len(roots), 2)
    eq_(set(roots), {foo, bar})

    roots = list(dig(foobar_foobar, cache={foo}))
    eq_(len(roots), 1)
    eq_(set(roots), {bar})

    myfoobar = Dependent("foobar", lambda foo1, foo2: foo1 + foo2,
                         depends_on=[foo, foo])
    roots = list(dig(foobar_foobar, context={myfoobar}))
    eq_(len(roots), 1)
    eq_(set(roots), {foo})


def test_draw():
    foo = Dependent("foo", lambda: "foo")
    bar = Dependent("bar", lambda foo: foo + "bar", dependencies=[foo])

    draw(bar)  # Does not throw an error
    draw(bar, cache={foo: "CACHED"})  # Does not throw an error


@raises(DependencyError)
def test_not_implemented_error():
    foo = Dependent("foo")
    solve(foo)



import pickle

from nose.tools import eq_

from .. import italian
from ...datasources import revision
from ...dependencies import solve
from .util import compare_extraction

BAD = [
    "anale",
    "ano",
    "bastardi", "bastardo",
    "buttana",
    "caca", "cacare", "cacata",
    "cacca", "cacche", "caccola", "caccole",
    "caga", "cagare", "cagata", "cagate", "cagato",
    "cavolata", "cavolate",
    "cazz", "cazzata", "cazzate", "cazzi", "cazzo", "cazzone", "cazzoni",
    "cesso",
    "ciccione",
    "ciuccia",
    "coglione", "coglioni",
    "cojone",
    "cretini", "cretino",
    "culo",
    "deficente", "deficenti",
    "fancazzista",
    "fanculo",
    "fica", "figa", "fighe", "figo",
    "fotte", "fottiti", "fottuto",
    "fregna",
    "froci", "frocio", "frocione",
    "gnocca",
    "idiota",
    "inculare", "inculato", "inculo",
    "kazzo",
    "maiala",
    "merd", "merda", "merdaccia", "merde", "merdosa", "merdoso",
    "mignotta",
    "minchia", "minkia",
    "pene", "peni",
    "pipi", "pippa", "pippe",
    "pirla",
    "piscio",
    "pisellino", "pisello", "pisellone",
    "pompini", "pompino",
    "porca", "porco",
    "porno",
    "pupu",
    "puttana", "puttane", "puttanella", "puttaniere",
    "puzza", "puzzano", "puzzate", "puzzava", "puzzi", "puzzolente", "puzzone",
    "ricchione",
    "sborra", "sburro",
    "scema", "scemi", "scemo",
    "schifo", "schifosa", "schifoso",
    "scopare", "scopata", "scopate", "scopato", "scopava",
    "scoreggia", "scorregge", "scorreggia",
    "seghe",
    "sfigati", "sfigato",
    "siffredi",
    "skifo",
    "sperma",
    "stocazzo",
    "stronza", "stronzata", "stronzate",
    "stronzi", "stronzo",
    "stupido",
    "suca",
    "succhia",
    "suka",
    "tette",
    "troia",
    "troie",
    "trombare",
    "vafanculo", "vaffanculo",
    "znccnk", "znccnl",
    "zoccola"
]

INFORMAL = [
    "ahah", "ahahah", "ahahaha", "ahahahah",
    "amo",
    "avete",
    "banana",
    "bla",
    "brutti", "brutto",
    "cavolo",
    "chiamo",
    "ciao", "ciaoo", "ciaooo", "ciaoooo",
    "cmq",
    "consigliamo",
    "corsivo",
    "dovete",
    "fermatevi",
    "frega",
    "grassetto",
    "hahaha", "hahahaha",
    "inserisci",
    "intestazione",
    "leggete",
    "lol",
    "mamma",
    "nascondi",
    "perche",
    "potete",
    "raga",
    "sapete",
    "scrivete",
    "scusate",
    "siete",
    "sto",
    "tua",
    "volete"
]

OTHER = [
    """
    Le locomotive del gruppo 851 erano un gruppo di locomotive a vapore delle
    Ferrovie dello Stato.

    Furono progettate e fatte costruire dalla Rete Adriatica (RA) quali
    macchine per il servizio di linea. Nel 1905, insieme alle locomotive dei
    gruppi poi FS 290, 600 e 870 anch'esse ex RA, vennero inserite tra quelle
    che le FS reputarono meritevoli di ulteriori commesse nell'attesa del
    completamento del progetto dei nuovi gruppi idonei a fronteggiare lo
    sviluppo del traffico conseguente alla statalizzazione.
    """
]


def test_badwords():
    compare_extraction(italian.revision.badwords_list, BAD, OTHER)


def test_informals():
    compare_extraction(italian.revision.informals_list, INFORMAL, OTHER)


def test_revision():
    # Words
    cache = {revision.text: "Furono progettate e m80 costruire dalla."}
    eq_(solve(italian.revision.words_list, cache=cache),
        ["Furono", "progettate", "e", "m80", "costruire", "dalla"])

    # Misspellings
    cache = {revision.text: 'Furono progettate e dalla worngly. <td>'}
    eq_(solve(italian.revision.misspellings_list, cache=cache), ["worngly"])

    # Infonoise
    cache = {revision.text: "Furono progettate e dalla!"}
    eq_(solve(italian.revision.infonoise, cache=cache), 7/22)


def test_pickling():

    eq_(italian, pickle.loads(pickle.dumps(italian)))



import pickle

from nose.tools import eq_

from .. import indonesian
from ...datasources import revision
from ...dependencies import solve
from .util import compare_extraction

BAD = [
    "anjing",
    "bajingan",
    "bangsat",
    "bispak",
    "bloon", "blo'on",
    "brengsek", "brengsex",
    "bencong",
    "babi",
    "cibai",
    "coley",
    "diselama",
    "escoduro",
    "fredrike",
    "fogh",
    "gauguin",
    "goblok",
    "gestapo",
    "husseins",
    "indon",
    "jambut",
    "jellinek",
    "jellygamat",
    "keparat",
    "kontol",
    "lonte",
    "loked",
    "lvmh",
    "malingsia",
    "memek",
    "monyong",
    "netnapa",
    "ngentot",
    "nesbitt",
    "panadta",
    "palaji",
    "perek",
    "pukimak",
    "pedofil",
    "riyhad",
    "sempak",
    "sinting",
    "steinway",
    "sukhano",
    "taenjamras",
    "tetek",
    "titit",
    "toket",
    "tzcesar",
    "thailaland",
    "thaicia"
]

INFORMAL = [
    "hai",
    "halo",
    "jancok"
]

OTHER = [
    """
    Gita Gutawa adalah seorang penyanyi sopran, aktris, dan penulis lagu
    berkebangsaan Indonesia. Ia adalah putri dari komponis Erwin Gutawa.
    Meskipun pada awalnya sempat belajar piano, Gita kemudian beralih ke vokal.
    Bakatnya ditemukan pada tahun 2004 saat ia sedang berlatih vokal,
    kemudian diminta untuk berduet dengan ADA Band.
    """
]


def test_badwords():
    compare_extraction(indonesian.revision.badwords_list, BAD, OTHER)


def test_informals():
    compare_extraction(indonesian.revision.informals_list, INFORMAL, OTHER)


def test_revision():
    # Words
    cache = {revision.text: "Gita Gutawa adalah seorang m80 sopran, aktris."}
    eq_(solve(indonesian.revision.words_list, cache=cache),
        ["Gita", "Gutawa", "adalah", "seorang", "m80", "sopran", "aktris"])

    # Misspellings
    cache = {revision.text: 'Setelah merilis album duet worngly. <td>'}
    eq_(solve(indonesian.revision.misspellings_list, cache=cache), ["worngly"])


def test_pickling():

    eq_(indonesian, pickle.loads(pickle.dumps(indonesian)))



"""
.. autoclass:: revscoring.extractors.api.APIExtractor
"""
import logging
from collections import defaultdict
from itertools import islice

import mwapi
from mwtypes import Namespace, Timestamp

from .. import dependencies
from ..datasources import (Datasource, RevisionMetadata, UserInfo,
                           parent_revision, revision, site, user)
from ..errors import RevisionNotFound
from .extractor import Extractor

logger = logging.getLogger(__name__)

revision_doc = Datasource("revision.doc")
parent_revision_doc = Datasource("parent_revision.doc")
previous_user_revision_doc = Datasource("previous_user_revision.doc")
page_creation_doc = Datasource("page_creation.doc")
user_doc = Datasource("user.doc")
site_doc = Datasource("site.doc")


class APIExtractor(Extractor):
    """
    Implements a :class:`revscoring.Extractor` using a
    MediaWiki API.

    :Parameters:
        session : :class:`mwapi.Session`
            An API session to use
        context : `dict` | `iterable`
            A collection of :class:`revscoring.Dependent` to
            inject when extracting.
        cache : `dict`
            A collection of pre-computed values to inject when extracting
    """
    def __init__(self, session, context=None, cache=None):
        cache = cache or {}
        context = dependencies.normalize_context(context)

        self.session = session

        local_cache = {
            site.namespace_map: self.get_namespace_map()
        }
        local_context = {d: d for d in [
            Datasource("revision.doc", self.process_revision_doc,
                       depends_on=[revision.id]),
            Datasource("revision.metadata", self.process_revision_metadata,
                       depends_on=[revision_doc]),
            Datasource("revision.text", self.process_revision_text,
                       depends_on=[revision_doc]),
            Datasource("parent_revision.doc", self.process_parent_revision_doc,
                       depends_on=[revision.metadata]),
            Datasource("parent_revision.metadata",
                       self.process_revision_metadata_if_exists,
                       depends_on=[parent_revision_doc]),
            Datasource("parent_revision.text",
                       self.process_revision_text,
                       depends_on=[parent_revision_doc]),
            Datasource("previous_user_revision.doc",
                       self.process_previous_user_revision_doc,
                       depends_on=[revision.metadata]),
            Datasource("previous_user_revision.metadata",
                       self.process_revision_metadata_if_exists,
                       depends_on=[previous_user_revision_doc]),
            Datasource("page_creation.doc",
                       self.process_page_creation_doc,
                       depends_on=[revision.metadata]),
            Datasource("page_creation.metadata",
                       self.process_revision_metadata,
                       depends_on=[page_creation_doc]),
            Datasource("user.doc",
                       self.process_user_doc,
                       depends_on=[revision.metadata]),
            Datasource("user.info",
                       self.process_user_info,
                       depends_on=[user_doc])
        ]}

        local_context.update(context)
        local_cache.update(cache)

        super().__init__(local_context, local_cache)

    def extract(self, rev_ids, dependents, context=None, caches=None):
        """
        Extracts a values for a set of
        :class:`~revscoring.dependents.dependent.Dependent` (e.g.
        :class:`~revscoring.features.feature.Feature` or
        :class:`~revscoring.datasources.datasource.Datasource`) for a revision
        or a set of revisions

        :Parameters:
            rev_ids : int | `iterable`
                Either a single rev_id or an `iterable` of rev_ids
            dependents : :class:`~revscoring.dependents.dependent.Dependent`
                A set of dependents to extract values for
            context : `dict` | `iterable`
                A set of call-specific
                :class:`~revscoring.dependents.dependent.Dependent` to inject
            cache : `dict`
                A set of call-specific pre-computed values to inject

        :Returns:
            An generator of extracted values if a single rev_id was provided or
            a genetator of (error, values) pairs where error is `None` if no
            error occured during extraction.
        """
        caches = caches or {}
        context = dependencies.normalize_context(context)

        if hasattr(rev_ids, "__iter__"):
            return self._extract_many(rev_ids, dependents, context, caches)
        else:
            rev_id = rev_ids
            cache = caches
            return self._extract(rev_id, dependents, context, cache)

    def _extract_many(self, rev_ids, dependents, context, caches):
        all_dependents = set(dependencies.expand(dependents))

        # Prime caches
        extract_caches = defaultdict(dict)
        for rev_id in caches:
            extract_caches[rev_id].update(caches[rev_id])

        # Build up caches for data that can be queried in batch
        if revision.metadata in all_dependents or \
           revision.text in all_dependents:
            rev_ids_missing_data = [
                rid for rid in rev_ids
                if rid not in extract_caches or
                revision.text not in extract_caches[rid] or
                revision.metadata not in extract_caches[rid] or
                revision_doc not in extract_caches[rid]
            ]
            rev_docs = self.get_rev_doc_map(rev_ids_missing_data)

            for rev_id in rev_ids:
                extract_caches[rev_id][revision_doc] = rev_docs.get(rev_id)

            if parent_revision.metadata in all_dependents or \
               parent_revision.text in all_dependents:

                parent_ids = [r.get('parentid') for r in rev_docs.values()
                              if r.get('parentid', 0) > 0]

                parent_rev_docs = self.get_rev_doc_map(parent_ids)

                for rev_doc in rev_docs.values():
                    extract_caches[rev_doc['revid']][parent_revision_doc] = \
                        parent_rev_docs.get(rev_doc['parentid'])

            if user.info in all_dependents:
                user_texts = [r.get('user') for r in rev_docs.values()]
                user_docs = self.get_user_doc_map(user_texts)

                for rev_doc in rev_docs.values():
                    extract_caches[rev_doc['revid']][user_doc] = \
                        user_docs.get(rev_doc.get('user'))

        # Now extract dependent values one-by-one
        for rev_id in rev_ids:
            try:
                values = self._extract(rev_id, dependents, context=context,
                                       cache=extract_caches[rev_id])
                yield None, list(values)
            except Exception as e:
                yield e, None

    def _extract(self, rev_id, dependents, cache, context):
        extract_cache = {revision.id: rev_id}
        extract_cache.update(cache)
        return self.solve(dependents, context=context, cache=extract_cache)

    def get_namespace_map(self):
        logger.info("Requesting site info from the API")
        doc = self.session.get(
            action="query", meta="siteinfo",
            siprop={'general', 'namespaces', 'namespacealiases'})

        return self.namespace_map_from_doc(doc['query'])

    def get_rev_doc_map(self, rev_ids, props={'ids', 'user', 'timestamp',
                                              'userid', 'comment', 'content',
                                              'flags', 'size'}):
        if len(rev_ids) == 0:
            return {}
        logger.info("Batch requesting {0} revisions from the API"
                    .format(len(rev_ids)))
        return {rd['revid']: rd
                for rd in self.query_revisions_by_revids(
                    revids=rev_ids,
                    rvprop=props)}

    def query_revisions_by_revids(self, revids, batch=50, **params):
        revids_iter = iter(revids)
        while True:
            batch_ids = list(islice(revids_iter, 0, batch))
            if len(batch_ids) == 0:
                break
            else:
                doc = self.session.get(action='query', prop='revisions',
                                       revids=batch_ids, **params)

                for page_doc in doc['query'].get('pages', {}).values():
                    page_meta = {k: v for k, v in page_doc.items()
                                 if k != 'revisions'}
                    if 'revisions' in page_doc:
                        for revision_doc in page_doc['revisions']:
                            revision_doc['page'] = page_meta
                            yield revision_doc

    def get_user_doc_map(self, user_texts, props={'blockinfo',
                                                  'implicitgroups',
                                                  'groups', 'registration',
                                                  'emailable', 'editcount',
                                                  'gender'}):
        if len(user_texts) == 0:
            return {}
        logger.info("Batch requesting {0} users from the API"
                    .format(len(user_texts)))
        return {ud['name']: ud
                for ud in self.query_users_by_text(user_texts, usprop=props)}

    def query_users_by_text(self, user_texts, batch=50, **params):
        user_texts_iter = iter(user_texts)
        while True:
            batch_texts = list(islice(user_texts_iter, 0, batch))
            if len(batch_texts) == 0:
                break
            else:
                doc = self.session.get(action='query', list='users',
                                       ususers=batch_texts, **params)

                for user_doc in doc['query'].get('users', []):
                    yield user_doc

    def process_revision_doc(self, rev_id):
        logger.info("Requesting a revision ({0}) from the API".format(rev_id))
        props = {'ids', 'user', 'timestamp', 'userid', 'comment',
                 'content', 'flags', 'size'}
        return self.get_rev_doc_map([rev_id], props=props).get(rev_id)

    def process_parent_revision_doc(self, revision_metadata):
        props = {'ids', 'user', 'timestamp', 'userid', 'comment',
                 'content', 'flags', 'size'}
        if revision_metadata.parent_id is not None and \
                revision_metadata.parent_id > 0:
            rev_id = revision_metadata.parent_id
            logger.info("Requesting a parent revision ({0}) from the API"
                        .format(rev_id))
            return self.get_rev_doc_map([rev_id], props=props).get(rev_id)
        else:
            return None

    def process_previous_user_revision_doc(self, revision_metadata):
        if revision_metadata.user_text is not None:
            logger.info("Requesting previous user revision ({0}) from the API"
                        .format(revision_metadata.user_text))
            doc = self.session.get(
                action="query",
                list="usercontribs",
                ucuser=revision_metadata.user_text,
                upprop={'ids', 'timestamp'},
                uclimit=1,
                ucdir="older",
                ucstart=str(revision_metadata.timestamp-1)
            )

            rev_docs = doc['query']['usercontribs']

            if len(rev_docs) > 0:
                return rev_docs[0]
            else:
                return None
        else:
            return None

    def process_page_creation_doc(self, revision_metadata):
        logger.info("Requesting page creation ({0}) from the API"
                    .format(revision_metadata.page_id))
        doc = self.session.get(
            action="query",
            prop="revisions",
            pageids=revision_metadata.page_id,
            rvdir="newer",
            rvlimit=1,
            rvprop={'ids', 'user', 'timestamp', 'userid', 'comment',
                    'flags', 'size'}
        )
        page_doc = doc['query'].get('pages', {'revisions': []}).values()
        rev_docs = page_doc['revisions']

        if len(rev_docs) == 1:
            return rev_docs[0]
        else:
            return None

    def process_user_doc(self, revision_metadata):
        logger.info("Requesting user info ({0}) from the API"
                    .format(revision_metadata.user_text))

        user_name = revision_metadata.user_text
        return self.get_user_doc_map([user_name]).get(user_name)

    @classmethod
    def process_revision_metadata(cls, revision_doc):
        if revision_doc is None:
            raise RevisionNotFound()
        return cls.revision_metadata_from_doc(revision_doc)

    @classmethod
    def process_revision_metadata_if_exists(cls, revision_doc):
        if revision_doc is None:
            return None
        else:
            return cls.revision_metadata_from_doc(revision_doc)

    @classmethod
    def process_user_info(cls, user_doc):
        return cls.user_info_from_doc(user_doc)

    @classmethod
    def process_revision_text(cls, revision_doc):
        if revision_doc is None:
            return None
        return revision_doc.get('*', "")

    @classmethod
    def revision_metadata_from_doc(cls, rev_doc):
        if rev_doc is None:
            return None
        try:
            timestamp = Timestamp(rev_doc.get('timestamp'))
        except ValueError:
            timestamp = None

        return RevisionMetadata(rev_doc.get('revid'),
                                rev_doc.get('parentid'),
                                rev_doc.get('user'),
                                rev_doc.get('userid'),
                                timestamp,
                                rev_doc.get('comment'),
                                rev_doc.get('page', {}).get('pageid'),
                                rev_doc.get('page', {}).get('ns'),
                                rev_doc.get('page', {}).get('title'),
                                rev_doc.get('size'),
                                'minor' in rev_doc)

    @classmethod
    def user_info_from_doc(cls, user_doc):
        if user_doc is None:
            return None
        try:
            registration = Timestamp(user_doc.get('registration'))
        except ValueError:
            registration = None

        return UserInfo(
            user_doc.get('userid'),
            user_doc.get('name'),
            user_doc.get('editcount'),
            registration,
            user_doc.get('groups', []),
            user_doc.get('implicitgroups', []),
            "emailable" in user_doc,
            user_doc.get('gender'),
            user_doc.get('blockid'),
            user_doc.get('blockedby'),
            user_doc.get('blockedbyid'),
            user_doc.get('blockedtimestamp'),
            user_doc.get('blockreason'),
            user_doc.get('blockexpiry')
        )

    @classmethod
    def namespace_map_from_doc(cls, site_doc):
        aliases = site_doc.get('namespacealiases', [])
        alias_map = {}
        for alias_doc in aliases:
            prev_list = alias_map.get(alias_doc['id'], [])
            prev_list.append(alias_doc['*'])
            alias_map[alias_doc['id']] = prev_list

        namespace_map = {}
        for ns_doc in site_doc.get('namespaces', {}).values():
            namespace = namespace_from_doc(ns_doc, aliases=alias_map)
            namespace_map[namespace.id] = namespace

        return namespace_map

    @classmethod
    def from_config(cls, config, name, section_key="extractors"):
        logger.info("Loading APIExtractor '{0}' from config.".format(name))
        section = config[section_key][name]
        kwargs = {k: v for k, v in section.items() if k != "class"}
        return cls(mwapi.Session(**kwargs))


def namespace_from_doc(doc, aliases=None):
    aliases = {}
    return Namespace(doc['id'],
                     doc['*'],
                     canonical=doc.get('canonical'),
                     aliases=set(aliases.get(doc['id'], [])),
                     case=doc['case'],
                     content='content' in doc)



"""
``revscoring score -h``
::

    Prints formatted information about a model file.


    Usage:
        module_info -h | --help
        module_info <model-file> [--as-json]

    Options:
        -h --help     Prints this documentation
        <model-file>  Path to a model file
        --as-json     Output model info as a JSON blob
"""
import json

import docopt

from ..scorer_models import ScorerModel


def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)
    scorer_model = ScorerModel.load(open(args['<model-file>'], 'rb'))
    as_json = args['--as-json']

    run(scorer_model, as_json)


def run(scorer_model, as_json):
    if not as_json:
        print(scorer_model.format_info())
    else:
        print(json.dumps(scorer_model.info()))



import re

from ...datasources import Datasource


class RegexExtractor(Datasource):

    def __init__(self, name, regexes, depends_on=None, if_none=None):
        super().__init__(name, self.process, depends_on=depends_on)
        group_pattern = r"\b(" + r"|".join(regexes) + r")\b"
        self.group_re = re.compile(group_pattern, flags=re.I)
        self.if_none = if_none

    def process(self, input):
        if input is None:
            if self.if_none is not None:
                self.if_none()
            else:
                return []
        else:
            return self._process(input)

    def _process(self, input):
        raise NotImplementedError()


class SegmentRegexExtractor(RegexExtractor):

    def __init__(self, name, segment_source, regexes, if_none=None):
        super().__init__(name, regexes, depends_on=[segment_source])

    def _process(self, segments):
        return [match.group(0)
                for segment in segments
                for match in self.group_re.finditer(segment)]


class TextRegexExtractor(RegexExtractor):

    def __init__(self, name, text_source, regexes, if_none=None):
        super().__init__(name, regexes, depends_on=[text_source])

    def _process(self, text):
        matches = [match.group(0)
                   for match in self.group_re.finditer(text)]
        return matches



import re

CATEGORY_RE = re.compile(r"^category:", re.U | re.I)
CITE_RE = re.compile(r"cite", re.I | re.U)
IMAGE_RE = re.compile(r"^(file|image):", re.U | re.I)
INFOBOX_RE = re.compile(r"infobox|sidebar", re.U | re.I)
MARKUP_RE = re.compile(r'(\[|\]|\{\||\|\}|\|-|\{\{|\}\})+')
NUMERIC_RE = re.compile(r'\d+', re.U)
SECTION_COMMENT_RE = re.compile(r"\/\*([^\*]|\*[^\/])+\*\/")
SYMBOLIC_RE = re.compile(r'[^\w0-9\s\n\r]+', re.U)
# UPPERCASE_RE = re.compile(r'[A-Z]+', re.UNICODE)



import io
import time
from datetime import datetime

from sklearn.metrics import auc, roc_curve
from tabulate import tabulate

from .scorer_model import MLScorerModel
from .util import normalize_json


class ScikitLearnClassifier(MLScorerModel):

    def __init__(self, features, classifier_model, version=None):
        super().__init__(features, version=version)
        self.classifier_model = classifier_model
        self.stats = None

    def __getattr__(self, attr):
        if attr is "stats":
            return None
        else:
            raise AttributeError(attr)

    def train(self, values_labels):
        """

        :Returns:
            A dictionary with the fields:

            * seconds_elapsed -- Time in seconds spent fitting the model
        """
        start = time.time()

        values, labels = zip(*values_labels)

        # Fit SVC model
        self.classifier_model.fit(values, labels)
        self.trained = time.time()

        return {
            'seconds_elapsed': time.time() - start
        }

    def score(self, feature_values):
        """
        Generates a score for a single revision based on a set of extracted
        feature_values.

        :Parameters:
            feature_values : collection(`mixed`)
                an ordered collection of values that correspond to the
                `Feature` s provided to the constructor

        :Returns:
            A dict with the fields:

            * predicion -- The most likely class
            * probability -- A mapping of probabilities for input classes
                             corresponding to the classes the classifier was
                             trained on.  Generating this probability is
                             slower than a simple prediction.
        """
        prediction = self.classifier_model.predict([feature_values])[0]
        labels = self.classifier_model.classes_
        probas = self.classifier_model.predict_proba([feature_values])[0]
        probability = {label: proba for label, proba in zip(labels, probas)}

        doc = {
            'prediction': prediction,
            'probability': probability
        }
        return normalize_json(doc)

    def test(self, values_labels):
        """
        :Returns:
            A dictionary of test statistics with the fields:

            * accuracy -- The mean accuracy of classification
            * table -- A truth table for classification
            * roc
                * auc -- The area under the ROC curve
        """
        values, labels = zip(*values_labels)

        scores = [self.score(feature_values) for feature_values in values]

        self.stats = {
            'table': self._label_table(scores, labels),
            'accuracy': self.classifier_model.score(values, labels),
            'roc': self._roc_stats(scores, labels,
                                   self.classifier_model.classes_)
        }
        return self.stats

    def info(self):
        return normalize_json({
            'type': self.__class__.__name__,
            'version': self.version,
            'trained': self.trained,
            'stats': self.stats
        })

    def format_info(self):
        info = self.info()
        formatted = io.StringIO()
        formatted.write("ScikitLearnClassifier\n")
        formatted.write(" - type: {0}\n".format(info.get('type')))
        formatted.write(" - version: {0}\n".format(info.get('version')))
        if isinstance(info['trained'], float):
            date_string = datetime.fromtimestamp(info['trained']).isoformat()
            formatted.write(" - trained: {0}\n".format(date_string))
        else:
            formatted.write(" - trained: {0}\n".format(info.get('trained')))

        formatted.write("\n")
        formatted.write(self.format_stats())
        return formatted.getvalue()

    def format_stats(self):
        if self.stats is None:
            return "No stats available"
        else:
            formatted = io.StringIO()
            predicted_actuals = self.stats['table'].keys()
            possible = list(set(actual for _, actual in predicted_actuals))
            possible.sort()

            formatted.write("Accuracy: {0}\n\n".format(self.stats['accuracy']))
            if 'auc' in self.stats['roc']:
                formatted.write("ROC-AUC: {0}\n\n"
                                .format(self.stats['roc']['auc']))
            else:
                formatted.write("ROC-AUC:\n")

                table_data = [[comparison_label,
                               self.stats['roc'][comparison_label]['auc']]
                              for comparison_label in possible]
                formatted.write(tabulate(table_data))
                formatted.write("\n\n")

            table_data = []

            for actual in possible:
                table_data.append(
                    [(str(actual))] +
                    [self.stats['table'].get((predicted, actual), 0)
                     for predicted in possible]
                )
            formatted.write(tabulate(
                table_data,
                headers=["~{0}".format(p) for p in possible]))

            return formatted.getvalue()

    @classmethod
    def _roc_stats(cls, scores, labels, possible_labels):

        if len(possible_labels) <= 2:
            # Binary classification, class choice doesn't matter.
            comparison_label = possible_labels[0]
            return cls._roc_single_class(scores, labels, comparison_label)
        else:
            roc_stats = {}
            for comparison_label in possible_labels:
                roc_stats[comparison_label] = \
                    cls._roc_single_class(scores, labels, comparison_label)

            return roc_stats

    @classmethod
    def _roc_single_class(cls, scores, labels, comparison_label):
        probabilities = [s['probability'][comparison_label]
                         for s in scores]

        true_positives = [l == comparison_label for l in labels]
        fpr, tpr, thresholds = roc_curve(true_positives, probabilities)

        return {
            'auc': auc(fpr, tpr)
        }

    @staticmethod
    def _label_table(scores, labels):

        predicteds = [s['prediction'] for s in scores]

        table = {}
        for pair in zip(labels, predicteds):
            table[pair] = table.get(pair, 0) + 1

        return table



from ...datasources import revision
from ...errors import RevisionNotFound
from ...features import Feature
from ..meta.infonoise import Infonoise
from ..meta.regex_extractors import TextRegexExtractor
from ..meta.token_filter import TokenFilter
from .util import token_is_word, utf16_cleanup


def raise_rnf():
    raise RevisionNotFound()


class Revision:
    """
    Implements a set of features based on the revision of interest.
    """

    DATASOURCE_MODULE = revision
    MODULE_NAME = "revision"

    def __init__(self, language, error_if_missing=False):
        self.language = language
        self.prefix = language.__name__ + "." + self.MODULE_NAME + "."

        self.words_list = TokenFilter(
            self.prefix + "words",
            self.DATASOURCE_MODULE.tokens,
            token_is_word,
            if_none=raise_rnf if error_if_missing else None
        )
        """
        Returns a list of word tokens.
        """

        self.words = Feature(
            self.prefix + "words", len,
            returns=int,
            depends_on=[self.words_list]
        )
        """
        A count of the number of words in the revision.
        """

        self.content_words_list = TokenFilter(
            self.prefix + "content_words",
            self.DATASOURCE_MODULE.content_tokens,
            token_is_word,
            if_none=raise_rnf if error_if_missing else None
        )
        """
        Returns a list of words that appear in the (non-markup) content of the
        revision.
        """

        self.content_words = Feature(
            self.prefix + "content_words", len,
            returns=int,
            depends_on=[self.content_words_list]
        )
        """
        A count of the number of words in the (non-markup) content of the
        revision.
        """

        if language.resources.stopwords is not None and \
           language.resources.stemmer is not None:
            self.infonoise = Infonoise(
                self.prefix + "infonoise",
                language.resources.stopwords,
                language.resources.stemmer.stem,
                self.content_words_list
            )
            """
            Returns a score measuring the proportion of text remaining after
            filtering markup and stopwords and stemming the remaining words.
            This feature is commonly used in quality prediction.
            """

        if language.resources.badwords is not None:
            self.badwords_list = TextRegexExtractor(
                self.prefix + "badwords",
                self.DATASOURCE_MODULE.text,
                language.resources.badwords,
                if_none=raise_rnf if error_if_missing else None
            )
            """
            Returns a list of the badwords that appear in the text.
            """

            self.badwords = Feature(
                self.prefix + "badwords", len,
                returns=int,
                depends_on=[self.badwords_list]
            )
            """
            Returns a count of the badwords that appear in the text.
            """

        if language.resources.informals is not None:
            self.informals_list = TextRegexExtractor(
                self.prefix + "informals",
                self.DATASOURCE_MODULE.text,
                language.resources.informals,
                if_none=raise_rnf if error_if_missing else None
            )
            """
            Returns a list of the informal words that appear in the text
            """

            self.informals = Feature(
                self.prefix + "informals", len,
                returns=int,
                depends_on=[self.informals_list]
            )
            """
            Returns a count of the informal words that appear in the text
            """

        if language.resources.dictionary is not None:
            self.misspellings_list = TokenFilter(
                self.prefix + "misspellings",
                self.words_list,
                self.not_in_dictionary
            )
            """
            Returns a list of the misspellings that appear in the text
            """

            self.misspellings = Feature(
                self.prefix + "misspellings", len,
                returns=int,
                depends_on=[self.misspellings_list]
            )
            """
            Returns a count of the misspellings that appear in the text
            """

    def in_dictionary(self, word):
        return self.language.resources.dictionary.check(utf16_cleanup(word))

    def not_in_dictionary(self, word):
        return not self.in_dictionary(word)



"""
This module implements a set of utilities for extracting features and
train/testing :class:`revscoring.scorer_models.MLScorerModel` from the
command-line.  When the revscoring python package is installed, a `revscoring`
utility should be available from the commandline.  Run `revscoring -h` for more
information:

extract_features
++++++++++++++++
.. automodule:: revscoring.utilities.extract_features

score
+++++
.. automodule:: revscoring.utilities.score

train_test
++++++++++
.. automodule:: revscoring.utilities.train_test
"""



import sys

from .space_delimited import SpaceDelimited

try:
    import enchant
    dictionary = enchant.Dict("es")
except enchant.errors.DictNotFoundError:
    raise ImportError("No enchant-compatible dictionary found for 'es'.  " +
                      "Consider installing 'myspell-es'.")

try:
    from nltk.stem.snowball import SnowballStemmer
    stemmer = SnowballStemmer("spanish")
except ValueError:
    raise ImportError("Could not load stemmer for {0}. ".format(__name__))

try:
    from nltk.corpus import stopwords as nltk_stopwords
    stopwords = set(nltk_stopwords.words('spanish'))
except LookupError:
    raise ImportError("Could not load stopwords for {0}. ".format(__name__) +
                      "You may need to install the nltk 'stopwords' " +
                      "corpora.  See http://www.nltk.org/data.html")

badwords = [
    r"awe(onao|vo)",  # idiot
    r"babosa",  # slug
    r"bastardo",  # bastard
    r"bollo",  # drunk (to the extreme)
    r"boludo",  # asshole
    r"bugarr[óo]n",  # gay/bi man
    r"buseta",  # ???
    r"caca", r"\w*cag(a|o|ar(ro)?)(n(es)?)?\w*",  # shit / shits / dirty
    r"caquita",  # shit
    r"cabr[óo]n(es)?",  # bastards
    r"cagad[ao]",  # funny / absurd
    r"capullo",  # end of a penis
    r"carajo", r"chinga(r|da)?", r"chingu?en",  # fuck
    r"chichornia",  # ???
    r"chiguero",  # ???
    r"chima(r|da|s)?",  # fuck
    r"chingad(o|azo|erita)s?",  # fuck
    r"chingo(n[cs]i[cs]imo|ner[íi]a|nes|rrón)s?",  # fuck
    r"chino(n[cs]i[cs]imo)?",  # chinese
    r"chosto",  # ???
    r"choch[oa]",  # cunt
    r"cholo",  # mixed race person
    r"chucha",  # fuck
    r"\w*chupa(n|r|mea?|mel[ao]|medias|pollas|ban?)?s?\w*",
    r"chupen(la|me|mela)", r"chupo",  # variants of suck / blow me
    r"cipote",  # child (slang)
    r"cochetumadre",  # car of your mother ???
    r"co[gj]er",  # have sex / fuck
    r"cojio",  # wedgie
    r"cojones?",  # balls
    r"comeme",  # "eat me"
    r"comian",  # ???
    r"conch(a|etumadre)",  # vagina (of your mother)
    r"conejo",  # rabbit ???
    r"consolador",  # dildo
    r"co[ñn]o",  # dammit
    r"cupan",  # Misspelling of "chupan" ???
    r"cuca",  # vagina
    r"cule(ar|ros?)",  # fuck
    r"cul(o|ito|iaos?)",  # ass
    r"cundango",  # gay / homosexual
    r"desmadr([ae]r?|ado|arse)?",  # chaotic ???
    r"dlaversh",  # ???
    r"droga(ta)?s?",  # drugs
    r"emput(e|ar|ado)s?",  # pissed off
    r"enc", r"encabron(ar|ado|adas|arse)s?",  # pissed off
    r"encronada",  # ???
    r"enputado",  # bitch / complain
    r"fach(a|ero|era)",  # superficial appearance
    r"foll(o|ar|aban?)?",  # fuck
    r"fornicar",  # sex
    r"freg[oó]n",  # wet blanket -- no fun ???
    r"furcia",  # whore
    r"gabacho",  # frenchy (disrespectful)
    r"gamberro",  # thug
    r"gilipoll[ao]s?",  # asshole
    r"gitan[ao]s?",  # gypsy
    r"gord[ao]s?",  # fat
    r"gringo",  # US national (historically derogatory)
    r"guev(o|on(es)?)",  # cock / idiot
    r"g[uü]ey",  # bitch ???
    r"guiri",  # foreigner / tourist (derogatory)
    r"güila",  # ???
    r"golfa",  # slut
    r"huevos",  # balls
    r"huev[óo]n(es)?", r"huevospateados?",  # lazy ass
    r"hifueputa",  # ???
    r"hijode?putas?", r"hijoduta",  # motherfucker
    r"hijueputas?",  # son of a bitch
    r"holocuento",  # the Hollocaust
    r"idiotas?", r"jidiotas?",  # idiot
    r"imb[ée][cs]il(es)?",  # imbecile
    r"japo",  # japanese (derogatory)
    r"jetear",  # sleep (slang)
    r"jili(puerta|polla)s?",  # cock sucker ???
    r"joder",  # fuck
    r"jot(a|o|eria)s?",  # faggot
    r"jud[íi]o",  # jew (derogatory)
    r"jod(er|ido)",  # fucked
    r"komekaka",  # misspelling of 'come caca' -- "eat shit"
    r"lesbiana",  # lesbian
    r"lamadre",  # 'your mother' explitive
    r"mach(orra|etorra)",  # lesbian (offensive)
    r"madrea(do|r)",  # to beat someone up
    r"maldito",  # dammit
    r"malparid[oa]s?",  # miscarried
    r"mama(da|guevo)s?",  # blowjob
    r"mam[óo]+n(es)?",  # insolent little douchebag
    r"manola",  # the (female) hand one masturbates with
    r"mari[ck](a|[óo]ne?)s?",  # gay man
    r"marimach[ao]",  # tomboy
    r"maripos[óo]n",  # fag
    r"marrano",  # pig
    r"meachupan", r"meapelan",  # something 'suck me' ???
    r"me[- ]?castra[- ]?la[- ]?madre",  # 'I <something> your mother' ???
    r"me[- ]?la[- ]?pelan?s?",  # "You ain't shit" ???
    r"\w*mi?er[dg]a+s?\w*",  # shit
    r"mimsn",  # ???
    r"ming[ao]",  # ugly person
    r"mocos",  # boogers / mucus
    r"mojon",  # shit
    r"monda",  # man part / penis
    r"moro",  # moor / muslim (derogatory)
    r"nacio",  # ???
    r"nazi",  # nazi
    r"negrat[oa]",  # nigger
    r"n[oe]m[ea]mes",  # ???
    r"ojetes?", "ogts?",  # asshole / stingy
    r"paj(a|aro|ero)s?",  # wank / wanker
    r"paki",  # ???
    r"pamearlo",  # ???
    r"pario",  # ???
    r"pattaya",  # ???
    r"pedos?",  # pedofile
    r"pel(ado|an)",  # boy (derogatory)
    r"pelot(a|udo)s?",  # balls
    r"pendej(o|a|ada)s?",  # stupid people
    r"penes?",  # penis
    r"peos?",  # fart
    r"perras?",  # bitch
    r"petard[oa]s?",  # slow / retarded
    r"pich(a|ula)",  # cock
    r"pijas?",  # cock
    r"piko",  # ???
    r"pinches?",  # kitchen boy / insignificant
    r"pinga",  # dick
    r"pipi",  # small penis
    r"pirobos",  # fag
    r"pit(o|ito|ote)s?",  # penis
    r"pollas?",  # cock
    r"poronga",  # cock
    r"polvo",  # nut (testical)
    r"poto",  # ass
    r"prostituta",  # prostitute
    r"\w*put([ao]+|isim[oa]|iza)(s|n)?\w*",  # bitch/whore
    r"puñal",  # fag ???
    r"rabo",  # ass
    r"ramera",  # easy woman
    r"sida",  # AIDS
    r"sorete",  # piece of shit
    r"sudaca",  # greaser
    r"s[íi]filis",  # syphilis
    r"tajodido",  # ???
    r"tetas?",  # tits
    r"tont[ao]s?",  # fool / low intelligence
    r"torta",  # fat woman
    r"tortillera",  # lesbian
    r"tranca",  # bar ???
    r"travesti",  # transvestite
    r"travolo",  # ???
    r"trol(o|a)",  # faggot
    r"uta", r"utama", r"utamadre",  # short for 'puta'
    r"\w*verg(a|uero|ud[ao])s?\w*", r"versh",  # cock
    r"vibrador",  # vibrator
    r"viol(o|a|e|ar)",  # violate
    r"vulva",  # vulva
    r"watdafuq",  # "what the fuck"
    r"wea",  # shit
    r"webon(ada|e)?s?",  # Chileans (derogatory)
    r"zapatona",  # shoe ???
    r"zorr(a|ear)"  # bitch
]
informals = [
    r"agan",  # ???
    r"agregenme",  # ???
    r"aguante",  # stamina
    r"aki",  # ???
    r"amaona",  # ???
    r"amigui",  # best (girl) friend
    r"amo+",  # love
    r"apesta",  # stinks
    r"asco",  # disgust
    r"asi",  # so
    r"atte?",  # short for atentamente
    r"(b+l+a+h*)+",
    r"bobada",  # ???
    r"bobos?",  # stupid / simple / fool
    r"bubis",  # ???
    r"chafa",  # low quality
    r"chale",  # no way!
    r"chido",  # cool
    r"chí",  # ???
    r"chil[ei]ar",  # ???
    r"comi[ao]",  # ???
    r"copien",  # copy
    r"cursiva",  # italics
    r"esq",  # abbrev. of 'esquina' meaning "corner" -- used in addresses
    r"est[uú]pid[aeo]+[rs]?",  # stupid
    r"fulan[ao]",  # pronoun for person
    r"fe[ao]s?",  # ugly
    r"gra(x|sias)",  # thank you
    r"guapo",  # beautiful
    r"h?o+la+",  # hello
    r"ho+lis?",  # ???
    r"ijos",  # ???
    r"inserta",  # inserts
    r"j+[eaiou]+(j+[aeiou]*)*",  # jaja jeje jojojo
    r"ke",  # short for 'que'
    r"kie(n|ro)",  # I want
    r"komo",  # ???
    r"lean",  # ???
    r"lees",  # ???
    r"lo[ck]os?",  # crazy
    r"(l+[uo]+l+)([uo]+l+)*",  # lol lollloololol lulllololul
    r"madrazo",  # a crash (slang)
    r"malparida",  # miscaried
    r"mcfinnigan",  # ???
    r"mensos?",  # stupid / annoying person
    r"meti[oa]",  # ???
    r"metroflog",  # ???
    r"migu?is",  # ???
    r"muxo",  # ???
    r"negrit[ao]",  # black person (non-offensive)
    r"nocheto",  # ???
    r"noo+",  # noooo
    r"nop",  # nope
    r"ojala",  # hopefully
    r"o+l[ia]",
    r"osea",  # "that is"
    r"pollid",  # ???
    r"popo",  # ???
    r"pipi",  # small penis
    r"plis",  # ???
    r"por[- ]?favor",  # please
    r"por[- ]?[kq]ue?", r"porke", r"porqe?",  # why!?
    r"porquer[ií]as?",  # low quality
    r"profe",  # slang for teacher
    r"pupu",  # ???
    r"qiero",  # ???
    r"salud(o?s)?",  # cheers!
    r"sierto",  # ???
    r"shí",  # ???
    r"sii+",  # "yessss"
    r"soi",  # I am
    r"sophonpanich",  # ???
    r"ta?mbn", r"tkm",  # also (slang)
    r"tanga",  # thong
    r"te[- ]?quiero[- ]?mucho", r"tqm",  # I care for you a lot
    r"umaxnet",  # ???
    r"vallanse",  # move out
    r"vayanse",  # go away
    r"wen[ao]",  # Hello
    r"weon(es)?",  # dude / thing
    r"wey",  # buddy
    r"xd+",  # XD (smiley face squiting)
    r"xfarm",  # ???
    r"yolo",  # You Only Live Once
    r"zorpia"  # ???
]

sys.modules[__name__] = SpaceDelimited(
    __name__,
    doc="""
spanish
=======

revision
--------
.. autoattribute:: revision.words
.. autoattribute:: revision.content_words
.. autoattribute:: revision.badwords
.. autoattribute:: revision.misspellings
.. autoattribute:: revision.informals
.. autoattribute:: revision.infonoise

parent_revision
---------------
.. autoattribute:: parent_revision.words
.. autoattribute:: parent_revision.content_words
.. autoattribute:: parent_revision.badwords
.. autoattribute:: parent_revision.misspellings
.. autoattribute:: parent_revision.informals
.. autoattribute:: parent_revision.infonoise

diff
----
.. autoattribute:: diff.words_added
.. autoattribute:: diff.words_removed
.. autoattribute:: diff.badwords_added
.. autoattribute:: diff.badwords_removed
.. autoattribute:: diff.informals_added
.. autoattribute:: diff.informals_removed
.. autoattribute:: diff.misspellings_added
.. autoattribute:: diff.misspellings_removed
    """,
    badwords=badwords,
    dictionary=dictionary,
    informals=informals,
    stemmer=stemmer,
    stopwords=stopwords
)
"""
spanish
"""



import pickle

from nose.tools import eq_

from .. import hebrew
from ...datasources import revision
from ...dependencies import solve
from .util import compare_extraction

BAD = [
    "שרמוטה"
]

INFORMAL = [
    "בגללך"  # Because of you
]

OTHER = [
    "בגלל", "חתול"
]


def test_badwords():
    compare_extraction(hebrew.revision.badwords_list, BAD, OTHER)


def test_informals():
    compare_extraction(hebrew.revision.informals_list, INFORMAL, OTHER)


def test_revision():
    # Words
    cache = {revision.text: "סוויפט גדלה בוויומיסינג, פנסילבניה, לנאשוויל"}
    eq_(solve(hebrew.revision.words_list, cache=cache),
        ["סוויפט", "גדלה", "בוויומיסינג", "פנסילבניה", "לנאשוויל"])

    # Misspellings
    cache = {revision.text: 'בגלל חטול <td>'}
    eq_(solve(hebrew.revision.misspellings_list, cache=cache), ["חטול"])


def test_pickling():

    eq_(hebrew, pickle.loads(pickle.dumps(hebrew)))



from itertools import groupby

from . import modifiers, parent_revision, revision
from ..datasources import diff
from .feature import Feature
from .util import MARKUP_RE, NUMERIC_RE, SYMBOLIC_RE

bytes_changed = revision.bytes - parent_revision.bytes
"""
Represents encoded content bytes changed in this edit. It uses UTF-8 encoding.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import parent_revision
        >>> list(extractor.extract(655097130, [diff.bytes_changed]))
        [297]
"""

bytes_changed_ratio = bytes_changed / modifiers.max(parent_revision.bytes, 1)
"""
Represents ratio of bytes changed in this edit compared to parent revision
size (in bytes).

:Returns:
    float

:Example:
    ..code-block:: python

        >>> from revscoring.features import parent_revision
        >>> list(extractor.extract(655097130, [diff.bytes_changed_ratio]))
        [0.012515275378197294]
"""


def process_segments_added(diff_added_segments):
    return len(diff_added_segments)

segments_added = Feature("diff.segments_added", process_segments_added,
                         returns=int, depends_on=[diff.added_segments])
"""
Represents number of segments added in this edit.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import diff
        >>> list(extractor.extract(655097130, [diff.segments_added]))
        [1]
"""


def process_segment_removed(diff_segments_removed):
    return len(diff_segments_removed)

segments_removed = Feature("segments_removed", process_segment_removed,
                           returns=int, depends_on=[diff.removed_segments])
"""
Represents number of segments removed in this edit.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import diff
        >>> list(extractor.extract(660405537, [diff.segments_removed]))
        [3]
"""


# ############################# Characters ####################################
def process_chars_added(diff_added_segments):
    return len("".join(diff_added_segments))

chars_added = Feature("diff.chars_added", process_chars_added,
                      returns=int, depends_on=[diff.added_segments])
"""
Represents number of characters added in this edit.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import diff
        >>> list(extractor.extract(655097130, [diff.chars_added]))
        [297]
"""


def process_chars_removed(diff_removed_segments):
    return len("".join(diff_removed_segments))

chars_removed = Feature("diff.chars_removed", process_chars_removed,
                        returns=int, depends_on=[diff.removed_segments])
"""
Represents number of characters removed in this edit.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import diff
        >>> list(extractor.extract(655097130, [diff.chars_removed]))
        [297]
"""

proportion_of_chars_removed = chars_removed / \
    modifiers.max(parent_revision.chars, 1)
"""
Represents ratio of characters removed in this edit compared to overall
characters in parent revision.

:Returns:
    float

:Example:
    ..code-block:: python

        >>> from revscoring.features import diff
        >>> list(
            extractor.extract(609079959, [diff.proportion_of_chars_removed]))
        [1.0]
"""
proportion_of_chars_added = chars_added / \
    modifiers.max(revision.chars, 1)
"""
Represents ratio of characters removed in this edit compared to overall
characters in revision.

:Returns:
    float

:Example:
    ..code-block:: python

        >>> from revscoring.features import diff
        >>> list(
            extractor.extract(655097130, [diff.proportion_of_chars_added]))
        [0.012366755496335776]
"""


def process_markup_chars_added(diff_added_segments):
    concat = "".join(diff_added_segments)
    return sum(len(m.group(0)) for m in MARKUP_RE.finditer(concat))

markup_chars_added = \
    Feature("diff.markup_chars_added", process_markup_chars_added,
            returns=int, depends_on=[diff.added_segments])
"""
Represents number of markup characters added in this edit.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import diff
        >>> list(extractor.extract(660405663, [diff.markup_chars_added]))
        [76]
"""


def process_markup_chars_removed(diff_removed_segments):
    concat = "".join(diff_removed_segments)
    return sum(len(m.group(0)) for m in MARKUP_RE.finditer(concat))

markup_chars_removed = \
    Feature("diff.markup_chars_removed", process_markup_chars_removed,
            returns=int, depends_on=[diff.removed_segments])
"""
Represents number of markup characters removed in this edit.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import diff
        >>> list(extractor.extract(656508079, [diff.markup_chars_removed]))
        [8]
"""


proportion_of_markup_chars_added = \
    markup_chars_added / modifiers.max(chars_added, 1)
"""
Represents ratio of markup characters added in this edit compared to overall
characters added in revision.

:Returns:
    float

:Example:
    ..code-block:: python

        >>> from revscoring.features import diff
        >>> list(extractor.extract(655097130,
        ...      [diff.proportion_of_markup_chars_added]))
        [0.013468013468013467]
"""


added_markup_chars_ratio = \
    proportion_of_markup_chars_added / \
    modifiers.max(parent_revision.proportion_of_markup_chars, 0.001)
"""
Represents ratio of added markup characters in this edit compared to added
markup characters in the last edit.

:Returns:
    float

:Example:
    ..code-block:: python

        >>> from revscoring.features import parent_revision
        >>> list(extractor.extract(655097130, [diff.added_markup_chars_ratio]))
        [0.45635401635401635]
"""


def process_numeric_chars_added(diff_added_segments):
    concat = "".join(diff_added_segments)
    return sum(len(m.group(0)) for m in NUMERIC_RE.finditer(concat))

numeric_chars_added = \
    Feature("diff.numeric_chars_added", process_numeric_chars_added,
            returns=int, depends_on=[diff.added_segments])
"""
Represents number of numeric characters added in this edit.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import diff
        >>> list(extractor.extract(660405663, [diff.numeric_chars_added]))
        [52]
"""


def process_numeric_chars_removed(diff_removed_segments):
    concat = "".join(diff_removed_segments)
    return sum(len(m.group(0)) for m in NUMERIC_RE.finditer(concat))

numeric_chars_removed = \
    Feature("diff.numeric_chars_removed", process_numeric_chars_removed,
            returns=int, depends_on=[diff.removed_segments])
"""
Represents number of numeric characters removed in this edit.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import diff
        >>> list(extractor.extract(656508079, [diff.numeric_chars_removed]))
        [6]
"""


proportion_of_numeric_chars_added = \
    numeric_chars_added / modifiers.max(chars_added, 1)
"""
Represents ratio of numeric characters added compared to all characters added
in this edit.

:Returns:
    float

:Example:
    ..code-block:: python

        >>> from revscoring.features import diff
        >>> list(extractor.extract(660405663,
        ...      [diff.proportion_of_numeric_chars_added]))
        [0.0363382250174703]
"""


added_number_chars_ratio = \
    proportion_of_numeric_chars_added / \
    modifiers.max(parent_revision.proportion_of_numeric_chars, 0.001)
"""
Represents ratio of added numeric characters in this edit compared to added
numeric characters in the last edit.

:Returns:
    float

:Example:
    ..code-block:: python

        >>> from revscoring.features import parent_revision
        >>> list(extractor.extract(660405663, [diff.added_number_chars_ratio]))
        [0.8469149674896002]
"""


def process_symbolic_chars_added(diff_added_segments):
    concat = "".join(diff_added_segments)
    return sum(len(m.group(0)) for m in SYMBOLIC_RE.finditer(concat))

symbolic_chars_added = \
    Feature("diff.symbolic_chars_added", process_symbolic_chars_added,
            returns=int, depends_on=[diff.added_segments])
"""
Represents number of symbolic characters added in this edit.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import diff
        >>> list(extractor.extract(655097130, [diff.symbolic_chars_added]))
        [20]
"""


def process_symbolic_chars_removed(diff_removed_segments):
    concat = "".join(diff_removed_segments)
    return sum(len(m.group(0)) for m in SYMBOLIC_RE.finditer(concat))

symbolic_chars_removed = \
    Feature("diff.symbolic_chars_removed", process_symbolic_chars_removed,
            returns=int, depends_on=[diff.removed_segments])
"""
Represents ratio of symbolic characters removed compared to all characters
added in this edit.

:Returns:
    float

:Example:
    ..code-block:: python

        >>> from revscoring.features import diff
        >>> list(extractor.extract(660405663,
        ...      [diff.proportion_of_symbolic_chars_added]))
        [0.13277428371767994]
"""


proportion_of_symbolic_chars_added = symbolic_chars_added / \
    modifiers.max(chars_added, 1)
"""
Represents ratio of symbolic characters added compared to overall characters
added in this edit.

:Returns:
    float

:Example:
    ..code-block:: python

        >>> from revscoring.features import diff
        >>> list(extractor.extract(
            655097130, [diff.proportion_of_symbolic_chars_added]))
        [0.06734006734006734]
"""
added_symbolic_chars_ratio = \
    proportion_of_symbolic_chars_added / \
    modifiers.max(parent_revision.proportion_of_symbolic_chars, 0.001)
"""
Represents ratio of added symbolic characters in this edit compared to added
symbolic characters in the last edit.

:Returns:
    float

:Example:
    ..code-block:: python

        >>> from revscoring.features import parent_revision
        >>> list(extractor.extract(655097130,
                 [diff.added_symbolic_chars_ratio]))
        [0.6290819445604794]
"""


def process_uppercase_chars_added(diff_added_segments):
    return sum((not c.lower() == c)
               for segment in diff_added_segments
               for c in segment)

uppercase_chars_added = \
    Feature("diff.uppercase_chars_added", process_uppercase_chars_added,
            returns=int, depends_on=[diff.added_segments])
"""
Represents number of uppercase characters added in this edit.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import diff
        >>> list(extractor.extract(655097130, [diff.uppercase_chars_added]))
        [9]
"""


def process_uppercase_chars_removed(diff_removed_segments):
    return sum((not c.lower() == c)
               for segment in diff_removed_segments
               for c in segment)

uppercase_chars_removed = \
    Feature("diff.uppercase_chars_removed", process_uppercase_chars_removed,
            returns=int, depends_on=[diff.removed_segments])
"""
Represents number of uppercase characters removed in this edit.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import diff
        >>> list(extractor.extract(659049823, [diff.uppercase_chars_removed]))
        [3]
"""


proportion_of_uppercase_chars_added = \
    uppercase_chars_added / modifiers.max(chars_added, 1)
"""
Represents ratio of uppercase characters added compared to overall characters
added in this edit.

:Returns:
    float

:Example:
    ..code-block:: python

        >>> from revscoring.features import diff
        >>> list(extractor.extract(
            655097130, [diff.proportion_of_uppercase_chars_added]))
        [0.030303030303030304]
"""


added_uppercase_chars_ratio = \
    proportion_of_uppercase_chars_added / \
    modifiers.max(parent_revision.proportion_of_uppercase_chars, 0.001)
"""
Represents ratio of uppercase characters added compared to overall uppercase
characters in parent revision.

:Returns:
    float

:Example:
    ..code-block:: python

        >>> from revscoring.features import diff
        >>> list(
            extractor.extract(655097130, [diff.added_uppercase_chars_ratio]))
        [0.030303030303030304]
"""


def process_longest_repeated_char_added(diff_added_segments):
    try:
        return max(sum(1 for _ in group)
                   for segment in diff_added_segments
                   for _, group in groupby(segment.lower()))
    except ValueError:
        # Happens when there's no segments added
        return 1

longest_repeated_char_added = \
    Feature("diff.longest_repeated_char_added",
            process_longest_repeated_char_added,
            returns=int, depends_on=[diff.added_segments])
"""
Represents number of the most repeated character added.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import diff
        >>> list(
            extractor.extract(655097130, [diff.longest_repeated_char_added]))
        [4]
"""


# ############################# tokens ########################################
def process_longest_token_added(diff_added_tokens):
    try:
        return max(len(token) for token in diff_added_tokens)
    except ValueError:
        # Happens when there's no tokens added
        return 1

longest_token_added = \
    Feature("diff.longest_token_added", process_longest_token_added,
            returns=int, depends_on=[diff.added_tokens])
"""
Represents length of the biggest token (e.g. word) added in this edit.

:Returns:
    int

:Example:
    ..code-block:: python

        >>> from revscoring.features import diff
        >>> list(extractor.extract(655097130, [diff.longest_token_added]))
        [12]
"""
